from os.path import exists, join

from bundlewrap.utils.testing import host_os, make_repo, run


def test_after(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "echo a >> " + join(str(tmpdir), "foo"),
                            'after': {"action:b"},
                        },
                        "b": {
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    with open(join(str(tmpdir), "foo")) as f:
        assert f.read() == "b\na\n"


def test_before(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "echo a >> " + join(str(tmpdir), "foo"),
                            'before': {"action:b"},
                        },
                        "b": {
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    with open(join(str(tmpdir), "foo")) as f:
        assert f.read() == "a\nb\n"


def test_before_fail_no_skip(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "false",
                            'before': {"action:b"},
                        },
                        "b": {
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 1
    with open(join(str(tmpdir), "foo")) as f:
        assert f.read() == "b\n"


def test_after_fail_no_skip(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "false",
                        },
                        "b": {
                            'after': {"action:a"},
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 1
    with open(join(str(tmpdir), "foo")) as f:
        assert f.read() == "b\n"


def test_before_fail_skip_with_needs(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "false",
                            'before': {"action:b"},
                        },
                        "b": {
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                            'needs': {"action:a"},
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 1
    assert b"skipped" in stdout
    assert not exists(join(str(tmpdir), "foo"))


def test_after_fail_skip_with_needed_by(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "false",
                            'needed_by': {"action:b"},
                        },
                        "b": {
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                            'after': {"action:a"},
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 1
    assert b"skipped" in stdout
    assert not exists(join(str(tmpdir), "foo"))


def test_chain_skip(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "false",
                        },
                        "b": {
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                            'needs': {"action:a"},
                        },
                        "c": {
                            'command': "echo c >> " + join(str(tmpdir), "foo"),
                            'after': {"action:b"},
                        },
                        "d": {
                            'command': "echo d >> " + join(str(tmpdir), "foo"),
                            'needs': {"action:b"},
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 1
    assert b"skipped" in stdout
    with open(join(str(tmpdir), "foo")) as f:
        assert f.read() == "c\n"


def test_chain_skip_no_cascade(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "a": {
                            'command': "false",
                        },
                        "b": {
                            'command': "echo b >> " + join(str(tmpdir), "foo"),
                            'needs': {"action:a"},
                            'cascade_skip': False,
                        },
                        "c": {
                            'command': "echo c >> " + join(str(tmpdir), "foo"),
                            'after': {"action:b"},
                        },
                        "d": {
                            'command': "echo d >> " + join(str(tmpdir), "foo"),
                            'needs': {"action:b"},
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 1
    assert b"skipped" in stdout
    with open(join(str(tmpdir), "foo")) as f:
        assert f.read() in ("c\nd\n", "d\nc\n")



from contextlib import suppress
from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items.pkg import Pkg, PKG_INSTALLED_LOCK
from bundlewrap.utils.text import mark_for_translation as _

class AptPkg(Pkg):
    """
    A package installed by apt.
    """
    BUNDLE_ATTRIBUTE_NAME = "pkg_apt"
    ITEM_TYPE_NAME = "pkg_apt"
    WHEN_CREATING_ATTRIBUTES = {
        'start_service': True,
    }
    _pkg_manual_cache = {}

    def pkg_all_installed(self):
        result = self.run("dpkg -l | grep '^ii'")
        for line in result.stdout.decode('utf-8').strip().split("\n"):
            pkg_name = line[4:].split()[0].replace(":", "_")
            yield "{}:{}".format(self.ITEM_TYPE_NAME, pkg_name)

    @property
    def expected_state(self):
        return {
            'installed': self.attributes['installed'],
            'mark': 'manual' if self.attributes['installed'] else None,
        }

    @property
    def actual_state(self):
        mark = None
        if self.attributes['installed']:
            if self.pkg_manually_installed():
                mark = 'manual'
            else:
                mark = 'auto'
        return {
            'installed': self.pkg_installed_cached(),
            'mark': mark,
        }

    def fix(self, status):
        with suppress(KeyError):
            self._pkg_install_cache.get(self.node.name, set()).remove(self.id)
        pkg_name = self.name.replace("_", ":")
        if 'installed' in status.keys_to_fix:
            if self.attributes['installed'] is False:
                self.pkg_remove()
            else:
                self._pkg_manual_cache.setdefault(self.node.name, set()).add(pkg_name)
                self.pkg_install()
        elif 'mark' in status.keys_to_fix:
            self._pkg_manual_cache.setdefault(self.node.name, set()).add(pkg_name)
            self.run("apt-mark manual {}".format(quote(pkg_name)))

    def pkg_install(self):
        runlevel = "" if self.when_creating['start_service'] else "RUNLEVEL=1 "
        self.run(
            runlevel +
            "DEBIAN_FRONTEND=noninteractive "
            "apt-get -qy -o Dpkg::Options::=--force-confold --no-install-recommends "
            "install {}".format(quote(self.name.replace("_", ":"))),
            may_fail=True,
        )

    def pkg_installed(self):
        result = self.run(
            "dpkg -s {} | grep '^Status: '".format(quote(self.name.replace("_", ":"))),
            may_fail=True,
        )
        return result.return_code == 0 and " installed" in result.stdout_text

    @staticmethod
    def pkg_in_cache(pkgid, cache):
        pkgtype, pkgname = pkgid.split(":")
        if "_" in pkgname:
            return pkgid in cache
        else:
            for cached_pkgid in cache:
                if cached_pkgid is None:
                    continue
                if cached_pkgid == pkgid or cached_pkgid.startswith(pkgid + ":"):
                    return True
            return False

    def pkg_remove(self):
        self.run(
            "DEBIAN_FRONTEND=noninteractive "
            "apt-get -qy --allow-remove-essential "
            "purge {}".format(quote(self.name.replace("_", ":")))
        )

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        super(AptPkg, cls).validate_attributes(bundle, item_id, attributes)

        if not isinstance(attributes.get('when_creating', {}).get('start_service', True), bool):
            raise BundleError(_(
                "expected boolean for 'start_service' on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))

    def pkg_manually_installed(self):
        # ensure we don't run `apt-mark showmanual` concurrently
        with PKG_INSTALLED_LOCK.lock(self.node.name):
            if self.node.name not in self._pkg_manual_cache:
                result = self.run("apt-mark showmanual")
                self._pkg_manual_cache[self.node.name] = set()
                for line in result.stdout.decode('utf-8').strip().splitlines():
                    self._pkg_manual_cache[self.node.name].add(line.strip())

        pkg_quoted = self.name.replace("_", ":")
        return pkg_quoted in self._pkg_manual_cache.get(self.node.name, set())



from re import compile

from ..exceptions import InvalidMagicStringException
from ..metadata import atomic
from . import error_context
from .dicts import _Atomic

MAGIC_STRINGS_PATTERN = compile(r"^!([a-zA-Z0-9_]+):(.+)$")


def convert_magic_strings(repo, obj):
    if not repo.magic_string_functions:
        # If we don't have any magic string functions, we just skip this
        # altogether. This eases migration from existing implementations
        # of magic strings to the builtin methods.
        return obj

    is_atomic = isinstance(obj, _Atomic)

    if isinstance(obj, str):
        m = MAGIC_STRINGS_PATTERN.match(obj)
        if m:
            func_name, func_args = m.groups()
            try:
                func = repo.magic_string_functions[func_name]
            except KeyError:
                raise InvalidMagicStringException(func_name)
            else:
                with error_context(magic_string=func_name):
                    obj = func(func_args)
    elif isinstance(obj, dict):
        obj = {k: convert_magic_strings(repo, v) for k, v in obj.items()}
    elif isinstance(obj, list):
        obj = [convert_magic_strings(repo, i) for i in obj]
    elif isinstance(obj, set):
        obj = {convert_magic_strings(repo, i) for i in obj}
    elif isinstance(obj, tuple):
        obj = tuple([convert_magic_strings(repo, i) for i in obj])
    if is_atomic:
        return atomic(obj)
    return obj



from shlex import quote

from bundlewrap.items.pkg import Pkg


class ApkPkg(Pkg):
    """
    A package installed by apk.
    """

    BUNDLE_ATTRIBUTE_NAME = "pkg_apk"
    ITEM_TYPE_NAME = "pkg_apk"

    @property
    def quoted(self):
        return quote(self.name)

    def pkg_all_installed(self):
        pkgs = self.run("apk list --installed").stdout.decode("utf-8")
        for line in pkgs.splitlines():
            pkg_name = line.split()[0]
            yield f"{self.ITEM_TYPE_NAME}:{pkg_name}"

    def pkg_install(self):
        self.run(f"apk add {self.quoted}", may_fail=True)

    def pkg_installed(self):
        result = self.run(f"apk info --installed {self.quoted}", may_fail=True)
        return result.return_code == 0 and self.quoted in result.stdout_text

    def pkg_remove(self):
        self.run(f"apk del {self.quoted}", may_fail=True)



from contextlib import suppress
from importlib.util import module_from_spec, spec_from_file_location
from inspect import isabstract
from os import listdir, mkdir, walk
from os.path import abspath, dirname, isdir, isfile, join
from sys import version_info

try:
    from tomllib import loads as toml_load
except ImportError:
    from rtoml import load as toml_load

VERSION_NEW_PACKAGING = (3, 10)
if version_info >= VERSION_NEW_PACKAGING:
    from importlib import metadata
    from packaging.requirements import Requirement
else:
    from pkg_resources import DistributionNotFound, require, VersionConflict  # needs setuptools

from . import items, VERSION, VERSION_STRING
from .bundle import FILENAME_ITEMS
from .exceptions import (
    NoSuchGroup,
    NoSuchNode,
    NoSuchTarget,
    NoSuchRepository,
    MissingRepoDependency,
    RepositoryError,
)
from .group import Group
from .metagen import MetadataGenerator
from .node import Node, NODE_ATTRS
from .secrets import FILENAME_SECRETS, generate_initial_secrets_cfg, SecretProxy
from .utils import (
    cached_property,
    error_context,
    get_file_contents,
    names,
)
from .utils.dicts import hash_state_dict
from .utils.scm import get_git_branch, get_git_clean, get_rev
from .utils.node_lambda import parallel_node_eval
from .utils.text import bold, mark_for_translation as _, red, validate_name
from .utils.ui import io

DIRNAME_BUNDLES = "bundles"
DIRNAME_DATA = "data"
DIRNAME_HOOKS = "hooks"
DIRNAME_ITEM_TYPES = "items"
DIRNAME_LIBS = "libs"
FILENAME_GROUPS = "groups.py"
FILENAME_MAGIC_STRINGS = "magic_strings.py"
FILENAME_NODES = "nodes.py"
FILENAME_REQUIREMENTS = "requirements.txt"

HOOK_EVENTS = (
    'action_run_end',
    'action_run_start',
    'apply_end',
    'apply_start',
    'item_apply_end',
    'item_apply_start',
    'lock_add',
    'lock_remove',
    'lock_show',
    'node_ssh_connect',
    'node_apply_end',
    'node_apply_start',
    'node_run_end',
    'node_run_start',
    'repo_init',
    'run_end',
    'run_start',
    'secret_key_use',
    'test',
    'test_node',
)

INITIAL_CONTENT = {
    FILENAME_GROUPS: _("""
#groups['group-1'] = {
#    'bundles': (
#        'bundle-1',
#    ),
#    'members': (
#        'node-1',
#    ),
#    'subgroups': (
#        'group-2',
#    ),
#}

groups['all'] = {
    'member_patterns': (
        r".*",
    ),
}
    """),

    FILENAME_NODES: _("""
nodes['node-1'] = {
    'hostname': "localhost",
}
    """),
    FILENAME_REQUIREMENTS: "bundlewrap>={}\n".format(VERSION_STRING),
    FILENAME_SECRETS: generate_initial_secrets_cfg,
}


def _check_requirements(lines):
    reqs = [Requirement(line) for line in lines]

    for req in reqs:
        # Ignore packages with a marker that does *not* apply to
        # the current system.
        if req.marker is not None and not req.marker.evaluate():
            continue

        for installed in metadata.distributions():
            if req.name.lower() == installed.name.lower():
                if req.specifier.contains(installed.version):
                    # We're good.
                    break
                else:
                    raise MissingRepoDependency(_(
                        "{x} Python package '{required}' is listed in {filename}, "
                        "but only '{existing_name}=={existing_version}' was found. "
                        "You probably have to upgrade it with `pip install {required}`."
                    ).format(
                        existing_name=installed.name,
                        existing_version=installed.version,
                        filename=FILENAME_REQUIREMENTS,
                        required=req,
                        x=red("!"),
                    ))
        else:
            raise MissingRepoDependency(_(
                "{x} Python package '{pkg}' is listed in {filename}, but wasn't found. "
                "You probably have to install it with `pip install {pkg}`."
            ).format(
                filename=FILENAME_REQUIREMENTS,
                pkg=req,
                x=red("!"),
            ))


# TODO Remove when dropping support for Python 3.9
def _check_requirements_legacy(lines):
    try:
        require(lines)
    except DistributionNotFound as exc:
        raise MissingRepoDependency(_(
            "{x} Python package '{pkg}' is listed in {filename}, but wasn't found. "
            "You probably have to install it with `pip install {pkg}`."
        ).format(
            filename=FILENAME_REQUIREMENTS,
            pkg=exc.req,
            x=red("!"),
        ))
    except VersionConflict as exc:
        raise MissingRepoDependency(_(
            "{x} Python package '{required}' is listed in {filename}, "
            "but only '{existing}' was found. "
            "You probably have to upgrade it with `pip install {required}`."
        ).format(
            existing=exc.dist,
            filename=FILENAME_REQUIREMENTS,
            required=exc.req,
            x=red("!"),
        ))


class HooksProxy:
    def __init__(self, repo, path):
        self.repo = repo
        self.__hook_cache = {}
        self.__module_cache = {}
        self.__path = path
        self.__registered_hooks = None

    def __getattr__(self, attrname):
        if attrname not in HOOK_EVENTS:
            raise AttributeError(attrname)

        if self.__registered_hooks is None:
            self._register_hooks()

        event = attrname

        if event not in self.__hook_cache:
            # build a list of files that define a hook for the event
            files = []
            for filename, events in self.__registered_hooks.items():
                if event in events:
                    files.append(filename)

            # define a function that calls all hook functions
            def hook(**kwargs):
                level_hint = _("repo-level hooks")
                if 'node' in kwargs:
                    level_hint = _("node-level hooks for node {}").format(kwargs['node'].name)
                elif 'nodes' in kwargs:
                    level_hint += _(" for {} nodes").format(len(kwargs['nodes']))

                for filename in files:
                    with io.job(_("{event}  Running {level_hint} from {filename}").format(
                        event=bold(event),
                        level_hint=level_hint,
                        filename=filename,
                    )):
                        with error_context(filename=filename):
                            self.__module_cache[filename][event](**kwargs)
            self.__hook_cache[event] = hook

        return self.__hook_cache[event]

    def _register_hooks(self):
        """
        Builds an internal dictionary of defined hooks.

        Priming __module_cache here is just a performance shortcut and
        could be left out.
        """
        self.__registered_hooks = {}

        if not isdir(self.__path):
            return

        for filename in listdir(self.__path):
            filepath = join(self.__path, filename)
            if not filename.endswith(".py") or \
                    not isfile(filepath) or \
                    filename.startswith("_"):
                continue
            self.__module_cache[filename] = {}
            self.__registered_hooks[filename] = []
            for name, obj in self.repo.get_all_attrs_from_file(filepath).items():
                if name not in HOOK_EVENTS:
                    continue
                self.__module_cache[filename][name] = obj
                self.__registered_hooks[filename].append(name)


class LibsProxy:
    def __init__(self, path):
        self.__module_cache = {}
        self.__path = path

    def __getattr__(self, attrname):
        if attrname.startswith("__") and attrname.endswith("__"):
            raise AttributeError(attrname)
        if attrname not in self.__module_cache:
            filename = attrname + ".py"
            filepath = join(self.__path, filename)
            try:
                spec = spec_from_file_location(
                    'bundlewrap.repo.libs_{}'.format(attrname),
                    filepath,
                )
                mod = module_from_spec(spec)
                spec.loader.exec_module(mod)
            except Exception:
                io.stderr(_("Exception while trying to load {}:").format(filepath))
                raise
            self.__module_cache[attrname] = mod
        return self.__module_cache[attrname]


class Repository(MetadataGenerator):
    def __init__(self, repo_path=None):
        super().__init__()

        if repo_path is None:
            self.path = "/dev/null"
        else:
            self.path = self._discover_root_path(abspath(repo_path))

        self._set_path(self.path)

        self.bundle_names = []
        self.group_dict = {}
        self.node_dict = {}
        self.node_attribute_functions = {}
        self.magic_string_functions = {}
        self._get_all_attr_code_cache = {}
        self._get_all_attr_result_cache = {}

        with io.job("Loading repository"):
            if repo_path is not None:
                self.populate_from_path(self.path)
            else:
                self.item_classes = list(self.items_from_dir(items.__path__[0]))

        self.hooks.repo_init(
            repo=self,
            version_string=VERSION_STRING,
            version_tuple=VERSION,
        )

    def __eq__(self, other):
        if self.path == "/dev/null":
            # in-memory repos are never equal
            return False
        return self.path == other.path

    def __repr__(self):
        return "<Repository at '{}'>".format(self.path)

    @staticmethod
    def is_repo(path):
        """
        Validates whether the given path is a bundlewrap repository.
        """
        try:
            assert isdir(path)
            assert isdir(join(path, DIRNAME_BUNDLES))
            assert isfile(join(path, FILENAME_NODES))
            assert isfile(join(path, FILENAME_GROUPS))
        except AssertionError:
            return False
        return True

    def add_group(self, group):
        """
        Adds the given group object to this repo.
        """
        if group.name in names(self.nodes):
            raise RepositoryError(_("you cannot have a node and a group "
                                    "both named '{}'").format(group.name))
        if group.name in names(self.groups):
            raise RepositoryError(_("you cannot have two groups "
                                    "both named '{}'").format(group.name))
        self.group_dict[group.name] = group

    def add_node(self, node):
        """
        Adds the given node object to this repo.
        """
        if node.name in names(self.groups):
            raise RepositoryError(_("you cannot have a node and a group "
                                    "both named '{}'").format(node.name))
        if node.name in names(self.nodes):
            raise RepositoryError(_("you cannot have two nodes "
                                    "both named '{}'").format(node.name))
        self.node_dict[node.name] = node

    @cached_property
    def branch(self):
        return get_git_branch()

    @cached_property
    def expected_state(self):
        repo_dict = {}
        for node in self.nodes:
            repo_dict[node.name] = node.hash()
        return repo_dict

    @cached_property
    def clean(self):
        return get_git_clean()

    @classmethod
    def create(cls, path):
        """
        Creates and returns a repository at path, which must exist and
        be empty.
        """
        if listdir(path):
            raise ValueError(_("'{}' is not an empty directory".format(
                path
            )))

        for filename, content in INITIAL_CONTENT.items():
            if callable(content):
                content = content()
            with open(join(path, filename), 'w') as f:
                f.write(content.strip() + "\n")

        mkdir(join(path, DIRNAME_BUNDLES))
        mkdir(join(path, DIRNAME_ITEM_TYPES))

        return cls(path)

    def create_bundle(self, bundle_name):
        """
        Creates an empty bundle.
        """
        if not validate_name(bundle_name):
            raise ValueError(_("'{}' is not a valid bundle name").format(bundle_name))

        bundle_dir = join(self.bundles_dir, bundle_name)

        # deliberately not using makedirs() so this will raise an
        # exception if the directory exists
        mkdir(bundle_dir)
        mkdir(join(bundle_dir, "files"))

        open(join(bundle_dir, FILENAME_ITEMS), 'a').close()

    def get_all_attrs_from_file(self, path, base_env=None):
        """
        Reads all 'attributes' (if it were a module) from a source file.
        """
        if base_env is None:
            base_env = {}

        if not base_env and path in self._get_all_attr_result_cache:
            # do not allow caching when passing in a base env because that
            # breaks repeated calls with different base envs for the same
            # file
            return self._get_all_attr_result_cache[path]

        if path not in self._get_all_attr_code_cache:
            source = get_file_contents(path)
            with error_context(path=path):
                self._get_all_attr_code_cache[path] = \
                    compile(source, path, mode='exec')

        code = self._get_all_attr_code_cache[path]
        env = base_env.copy()
        with error_context(path=path):
            exec(code, env)

        if not base_env:
            self._get_all_attr_result_cache[path] = env

        return env

    def nodes_or_groups_from_file(self, path, attribute, preexisting):
        def node_attribute(func):
            if func.__name__ in NODE_ATTRS:
                raise RepositoryError(_(
                    "cannot redefine builtin attribute '{attr}' as @node_attribute in nodes.py"
                ).format(attr=func.__name__))
            self.node_attribute_functions[func.__name__] = func
            return func

        try:
            flat_dict = self.get_all_attrs_from_file(
                path,
                base_env={
                    attribute: preexisting,
                    'libs': self.libs,
                    'node_attribute': node_attribute,
                    'repo_path': self.path,
                    'vault': self.vault,
                },
            )[attribute]
        except KeyError:
            raise RepositoryError(_(
                "{} must define a '{}' variable"
            ).format(path, attribute))
        if not isinstance(flat_dict, dict):
            raise ValueError(_("'{v}' in '{p}' must be a dict").format(
                v=attribute,
                p=path,
            ))
        for name, infodict in flat_dict.items():
            infodict.setdefault('file_path', path)
            yield (name, infodict)

    def nodes_or_groups_from_dir(self, directory):
        path = join(self.path, directory)
        if not isdir(path):
            return {}
        result = {}
        for root_dir, _dirs, files in walk(path):
            for filename in files:
                filepath = join(root_dir, filename)
                if not filename.endswith(".toml") or \
                        not isfile(filepath) or \
                        filename.startswith("_"):
                    continue
                entity_name = filename[:-5]
                if entity_name in result:
                    raise RepositoryError(_(
                        "Duplicate definition of {entity_name} in {file1} and {file2}"
                    ).format(
                        entity_name=entity_name,
                        file1=filepath,
                        file2=result[entity_name]['file_path'],
                    ))
                with error_context(filepath=filepath):
                    infodict = toml_load(get_file_contents(filepath).decode())
                infodict['file_path'] = filepath
                result[entity_name] = infodict
        return result

    def get_magic_strings(self):
        if not isfile(self.magic_strings_file):
            return

        def magic_string(func):
            self.magic_string_functions[func.__name__] = func
            return func

        # We do not store the gotten attrs anywhere, because we're
        # only interested in the defined magic strings.
        self.get_all_attrs_from_file(
            self.magic_strings_file,
            base_env={
                'libs': self.libs,
                'magic_string': magic_string,
                'repo_path': self.path,
                'vault': self.vault,
            },
        )

    def items_from_dir(self, path):
        """
        Looks for Item subclasses in the given path.

        An alternative method would involve metaclasses (as Django
        does it), but then it gets very hard to have two separate repos
        in the same process, because both of them would register config
        item classes globally.
        """
        if not isdir(path):
            return
        for root_dir, _dirs, files in walk(path):
            for filename in files:
                filepath = join(root_dir, filename)
                if not filename.endswith(".py") or \
                        not isfile(filepath) or \
                        filename.startswith("_"):
                    continue
                for name, obj in self.get_all_attrs_from_file(filepath).items():
                    if obj == items.Item or name.startswith("_"):
                        continue
                    with suppress(TypeError):
                        if issubclass(obj, items.Item) and not isabstract(obj):
                            yield obj

    def _discover_root_path(self, path):
        while True:
            if self.is_repo(path):
                return path

            previous_component = dirname(path)
            if path == previous_component:
                raise NoSuchRepository

            path = previous_component

    def get_group(self, group_name):
        try:
            return self.group_dict[group_name]
        except KeyError:
            raise NoSuchGroup(group_name)

    def get_node(self, node_name):
        try:
            return self.node_dict[node_name]
        except KeyError:
            raise NoSuchNode(node_name)

    def group_membership_hash(self):
        return hash_state_dict(sorted(names(self.groups)))

    @property
    def groups(self):
        return set(self.group_dict.values())

    def hash(self):
        return hash_state_dict(self.expected_state)

    @property
    def nodes(self):
        return set(self.node_dict.values())

    def nodes_in_all_groups(self, group_names):
        """
        Returns a list of nodes where every node is a member of every
        group given.

        :param group_names: list of names of the groups to check for
        :return list of nodes where every node is a member of every group given.
        """
        base_group = set(self.get_group(group_names[0]).nodes)
        for group_name in group_names[1:]:
            if not base_group:
                # quit early if we have already eliminated every node
                break
            base_group.intersection_update(set(self.get_group(group_name).nodes))
        result = list(base_group)
        result.sort()
        return result

    def nodes_in_any_group(self, group_names):
        """
        Returns all nodes that are a member of at least one of the given
        groups.

        :param group_names: list of names of the groups to check for
        :return list of nodes which are in at least one of given groups
        """
        for node in self.nodes:
            if node.in_any_group(group_names):
                yield node

    def nodes_in_group(self, group_name):
        """
        Returns a list of nodes in the given group.

        :param group_name: name of the group to check for
        :return list of nodes which are in the given group
        """
        return self.nodes_in_all_groups([group_name])

    def nodes_not_in_group(self, group_name):
        """
        Returns a list of nodes not in the given group.

        :param group_name: name of the group to check for
        :return list of nodes which are not in the given group
        """
        return [
            node
            for node in self.nodes
            if not node.in_group(group_name)
        ]

    def nodes_with_bundle(self, bundle_name):
        """
        Returns a list of nodes that do have the given bundle.

        :param bundle_name: name of the bundle to check for
        :return list of nodes which have the given bundle
        """
        return [
            node
            for node in self.nodes
            if bundle_name in names(node.bundles)
        ]

    def nodes_without_bundle(self, bundle_name):
        """
        Returns a list of nodes that do not have the given bundle.

        :param bundle_name: name of the bundle to check for
        :return list of nodes which do not have the given bundle
        """
        return [
            node
            for node in self.nodes
            if bundle_name not in names(node.bundles)
        ]

    def nodes_matching_lambda(self, lambda_str, lambda_workers=None):
        """
        Returns a list of nodes matching the lambda.

        Example:
            nodes = repo.nodes_matching_lambda("lambda:node.metadata_get('foo/magic', 47) < 3")

        :param lambda_str: string to evaluate as python code with `node` being one of the nodes,
            expected to return a value that can be interpreted as boolean
        :param lambda_workers: number of parallel workers used to check lambda condition on every node
        :return list of nodes matching the given lambda
        """
        result_items = parallel_node_eval(
            self.nodes,
            lambda_str,
            lambda_workers,
        ).items()

        return [
            self.get_node(node_name)
            for node_name, result in result_items
            if result
        ]

    def nodes_matching(self, target_strings, lambda_workers=None):
        """
        Returns a list of nodes matching any of the given target-strings. This is the same API that is used by
        all the bw commandlines, i.e. `bw items` or `bw apply` to select which nodes to operate on.

        Example:
            nodes = repo.nodes_matching(['wi-5.s2s', 'wi-5.routing'])

        :param target_strings: expression to select target nodes:
        my_node            # to select a single node
        my_group           # all nodes in this group
        bundle:my_bundle   # all nodes with this bundle
        !bundle:my_bundle  # all nodes without this bundle
        !group:my_group    # all nodes not in this group
        "lambda:node.metadata_get('foo/magic', 47) < 3"
        # all nodes whose metadata.get("foo/magic") is less than three

        :param lambda_workers: number of parallel workers to check lambda condition on nodes
        :return list of nodes matching any of the given target-strings
        """
        if isinstance(target_strings, str):
            target_strings = [target_strings]

        targets = set()
        for name in target_strings:
            name = name.strip()
            if name.startswith("bundle:"):
                bundle_name = name.split(":", 1)[1]
                targets.update(self.nodes_with_bundle(bundle_name))
            elif name.startswith("!bundle:"):
                bundle_name = name.split(":", 1)[1]
                targets.update(self.nodes_without_bundle(bundle_name))
            elif name.startswith("!group:"):
                group_name = name.split(":", 1)[1]
                targets.update(self.nodes_not_in_group(group_name))
            elif name.startswith("lambda:"):
                lambda_str = name.split(":", 1)[1]
                targets.update(self.nodes_matching_lambda(lambda_str, lambda_workers))
            else:
                try:
                    targets.add(self.get_node(name))
                except NoSuchNode:
                    try:
                        group = self.get_group(name)
                        targets.update(group.nodes)
                    except NoSuchGroup:
                        raise NoSuchTarget(name)

        return list(targets)

    def metadata_hash(self):
        repo_dict = {}
        for node in self.nodes:
            repo_dict[node.name] = node.metadata_hash()
        return hash_state_dict(repo_dict)

    def populate_from_path(self, path):
        if not self.is_repo(path):
            raise NoSuchRepository(
                _("'{}' is not a bundlewrap repository").format(path)
            )

        if path != self.path:
            self._set_path(path)

        # check requirements.txt
        try:
            with open(join(path, FILENAME_REQUIREMENTS)) as f:
                lines = [line.strip() for line in f.readlines()]

                # Ignore empty lines and comments.
                lines = [line for line in lines if line and not line.startswith('#')]

                # "-e some/editable" and "-r other_requirements.txt" are not
                # supported.
                lines = [line for line in lines if not line.startswith('-')]
        except Exception:
            pass
        else:
            if version_info >= VERSION_NEW_PACKAGING:
                _check_requirements(lines)
            else:
                _check_requirements_legacy(lines)

        self.vault = SecretProxy(self)

        # populate bundles
        self.bundle_names = []
        for dir_entry in listdir(self.bundles_dir):
            if validate_name(dir_entry):
                self.bundle_names.append(dir_entry)

        # populate magic strings
        self.get_magic_strings()

        # populate groups
        toml_groups = self.nodes_or_groups_from_dir("groups")
        self.group_dict = {}
        for group_name, group_attrs in self.nodes_or_groups_from_file(self.groups_file, 'groups', toml_groups):
            self.add_group(Group(group_name, attributes=group_attrs, repo=self))

        # populate items
        self.item_classes = list(self.items_from_dir(items.__path__[0]))
        for item_class in self.items_from_dir(self.items_dir):
            self.item_classes.append(item_class)

        # populate nodes
        toml_nodes = self.nodes_or_groups_from_dir("nodes")
        self.node_dict = {}
        for node_name, node_attrs in self.nodes_or_groups_from_file(self.nodes_file, 'nodes', toml_nodes):
            self.add_node(Node(node_name, attributes=node_attrs, repo=self))

    @cached_property
    def revision(self):
        return get_rev()

    def _set_path(self, path):
        self.path = path
        self.bundles_dir = join(self.path, DIRNAME_BUNDLES)
        self.data_dir = join(self.path, DIRNAME_DATA)
        self.hooks_dir = join(self.path, DIRNAME_HOOKS)
        self.items_dir = join(self.path, DIRNAME_ITEM_TYPES)
        self.groups_file = join(self.path, FILENAME_GROUPS)
        self.libs_dir = join(self.path, DIRNAME_LIBS)
        self.magic_strings_file = join(self.path, FILENAME_MAGIC_STRINGS)
        self.nodes_file = join(self.path, FILENAME_NODES)

        self.hooks = HooksProxy(self, self.hooks_dir)
        self.libs = LibsProxy(self.libs_dir)



from ..concurrency import WorkerPool
from ..utils.cmdline import get_target_nodes
from ..utils.table import ROW_SEPARATOR, render_table
from ..utils.text import blue, bold, error_summary, format_duration, green
from ..utils.text import mark_for_translation as _
from ..utils.text import red, red_unless_zero, yellow
from ..utils.ui import io


def bw_ipmi(repo, args):
    errors = []
    target_nodes = get_target_nodes(repo, args['targets'], args['node_workers'])
    pending_nodes = target_nodes.copy()
    io.progress_set_total(len(pending_nodes))

    results = {}

    def tasks_available():
        return bool(pending_nodes)

    def next_task():
        node = pending_nodes.pop()
        return {
            'target': node.run_ipmitool,
            'task_id': node.name,
            'kwargs': {
                'command': args['command'],
                'log_output': True,
            },
        }

    def handle_result(task_id, return_value, duration):
        io.progress_advance()
        results[task_id] = return_value

    def handle_exception(task_id, exception, traceback):
        io.progress_advance()
        msg = "{}  {}".format(bold(task_id), exception)
        io.stderr(traceback)
        io.stderr(repr(exception))
        io.stderr("{} {}".format(red("!"), msg))
        errors.append(msg)

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_result=handle_result,
        handle_exception=handle_exception,
        pool_id="ipmi",
        workers=args['node_workers'],
    )
    worker_pool.run()

    stats_table = [
        [bold(_("node")), bold(_("return code")), bold(_("time"))],
        ROW_SEPARATOR,
    ]

    for node_name, result in sorted(results.items()):
        if result is None:
            continue
        elif result.return_code == 0:
            return_code = green(0)
        else:
            return_code = red(result.return_code)

        stats_table.append([
            node_name,
            return_code,
            format_duration(result.duration),
        ])

    for line in render_table(stats_table):
        io.stdout("{x} {line}".format(x=blue("i"), line=line))
    error_summary(errors)

    exit(1 if errors else 0)



from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import mark_for_translation as _


def svc_start(node, svcname):
    return node.run(f"rc-service {quote(svcname)} start", may_fail=True)


def svc_running(node, svcname):
    result = node.run(f"rc-service {quote(svcname)} status", may_fail=True)
    return result.return_code == 0 and "started" in result.stdout_text


def svc_stop(node, svcname):
    return node.run(f"rc-service {quote(svcname)} stop", may_fail=True)


def svc_enable(node, svcname, runlevel):
    return node.run(f"rc-update add {quote(svcname)} {quote(runlevel)}", may_fail=True)


def svc_enabled(node, svcname, runlevel):
    result = node.run(
        f"rc-update show {quote(runlevel)} | grep -w {quote(svcname)}", may_fail=True
    )
    return result.return_code == 0 and svcname in result.stdout_text


def svc_runlevel(node, svcname):
    result = node.run(
        f"rc-update show --all | grep -w {quote(svcname)} | cut -d \\| -f 2",
        may_fail=True,
    )
    return result.stdout_text.strip() if result.return_code == 0 else None


def svc_disable(node, svcname, runlevel):
    return node.run(f"rc-update del {quote(svcname)} {quote(runlevel)}", may_fail=True)


class SvcOpenRC(Item):
    """
    A service managed by OpenRC init scripts.
    """

    BUNDLE_ATTRIBUTE_NAME = "svc_openrc"
    ITEM_ATTRIBUTES = {
        "running": True,
        "enabled": True,
        "runlevel": "default",
    }
    ITEM_TYPE_NAME = "svc_openrc"

    def __repr__(self):
        return "<SvcOpenRC name:{} runlevel:{} enabled:{} running:{}>".format(
            self.name, self.attributes["runlevel"], self.attributes["enabled"], self.attributes["running"],
        )

    def fix(self, status):
        if "enabled" in status.keys_to_fix:
            if self.attributes["enabled"]:
                svc_enable(self.node, self.name, self.attributes["runlevel"])
            else:
                svc_disable(self.node, self.name, self.attributes["runlevel"])

        if "running" in status.keys_to_fix:
            if self.attributes["running"]:
                svc_start(self.node, self.name)
            else:
                svc_stop(self.node, self.name)

        if "runlevel" in status.keys_to_fix:
            if status.actual_state.runlevel:
                svc_disable(self.node, self.name, status.actual_state.runlevel)
            svc_enable(self.node, self.name, self.attributes["runlevel"])

    def get_canned_actions(self):
        return {
            "stop": {
                "command": f"rc-service {self.name} stop",
                "needed_by": {self.id},
            },
            "restart": {
                "command": f"rc-service {self.name} restart",
                "needs": {self.id},
            },
            "reload": {
                "command": f"rc-service {self.name} reload".format(self.name),
                "needs": {
                    # make sure we don't reload and restart simultaneously
                    f"{self.id}:restart",
                    # with only the dep on restart, we might still end
                    # up reloading if the service itself is skipped
                    # because the restart action has cascade_skip False
                    self.id,
                },
            },
        }

    @property
    def actual_state(self):
        return {
            "enabled": svc_enabled(self.node, self.name, self.attributes["runlevel"]),
            "running": svc_running(self.node, self.name),
            "runlevel": svc_runlevel(self.node, self.name),
        }

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        for attribute in ("enabled", "running"):
            if attributes.get(attribute, None) not in (True, False, None):
                raise BundleError(
                    _(
                        "expected boolean or None for '{attribute}' on {item} in bundle '{bundle}'"
                    ).format(
                        attribute=attribute, bundle=bundle.name, item=item_id,
                    )
                )



from collections import Counter
from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import mark_for_translation as _


class ZFSPool(Item):
    """
    Creates ZFS pools.
    """
    BUNDLE_ATTRIBUTE_NAME = "zfs_pools"
    ITEM_ATTRIBUTES = {
        'autoexpand': None,
        'autoreplace': None,
        'autotrim': None,
    }
    WHEN_CREATING_ATTRIBUTES = {
        'ashift': None,
        'config': None,
    }
    ITEM_TYPE_NAME = "zfs_pool"

    def __repr__(self):
        return "<ZFSPool name:{} autoexpand:{} autoreplace:{} autotrim:{} ashift:{} config:{}>".format(
            self.name,
            self.attributes['autoexpand'],
            self.attributes['autoreplace'],
            self.attributes['autotrim'],
            self.when_creating['ashift'],
            self.when_creating['config'],
        )

    @property
    def expected_state(self):
        state = {}
        for i in self.attributes:
            if self.attributes.get(i) is not None:
                state[i] = self.attributes[i]

        return state

    @property
    def devices_used(self):
        devices = set()
        for option in self.when_creating['config']:
            for device in option['devices']:
                devices.add(device)
        return sorted(devices)

    def fix(self, status):
        if status.must_be_created:
            cmdline = []
            for option in self.when_creating['config']:
                if option.get('type'):
                    cmdline.append(option['type'])
                    if option['type'] == 'log' and len(option['devices']) > 1:
                        cmdline.append('mirror')

                for device in sorted(option['devices']):
                    res = self.run("lsblk -rndo fstype {}".format(quote(device)))
                    detected = res.stdout.decode('UTF-8').strip()
                    if detected != "":
                        raise BundleError(_(
                            "Node {}, ZFSPool {}: Device {} to be used for ZFS, "
                            "but it is not empty! Has '{}'."
                        ).format(self.node.name, self.name, device, detected))

                    cmdline.append(quote(device))

            options = set()
            if self.when_creating['ashift']:
                options.add('-o ashift={}'.format(self.when_creating['ashift']))

            for opt, value in status.expected_state.items():
                state_str = 'on' if value else 'off'
                options.add('-o {}={}'.format(opt, state_str))

            self.run('zpool create {} {} {}'.format(
                ' '.join(sorted(options)),
                quote(self.name),
                ' '.join(cmdline),
            ))
        elif status.keys_to_fix:
            for attr in status.keys_to_fix:
                state_str = 'on' if status.expected_state[attr] else 'off'
                self.run('zpool set {}={} {}'.format(attr, state_str, quote(self.name)))

    @property
    def actual_state(self):
        status_result = self.run('zpool list {}'.format(quote(self.name)), may_fail=True)
        if status_result.return_code != 0:
            return None

        pool_status = {}
        for line in self.run(
            'zpool get all -H -o all {}'.format(quote(self.name)),
            may_fail=True,
        ).stdout.decode().splitlines():
            try:
                pname, prop, value, source = line.split()
                pool_status[prop.strip()] = value.strip()
            except (IndexError, ValueError):
                continue

        state = {}
        for attr in self.attributes:
            state[attr] = (pool_status.get(attr) == 'on')

        return state

    def test(self):
        duplicate_devices = [
            item for item, count in Counter(self.devices_used).items() if count > 1
        ]
        if duplicate_devices:
            raise BundleError(_(
                "{item} on node {node} uses {devices} more than once as an underlying device"
            ).format(
                item=self.id,
                node=self.node.name,
                devices=_(" and ").join(duplicate_devices),
            ))

        # Have a look at all other ZFS pools on this node and check if
        # multiple pools try to use the same device.
        for item in self.node.items:
            if (
                item.ITEM_TYPE_NAME == "zfs_pool" and
                item.name != self.name and
                set(item.devices_used).intersection(set(self.devices_used))
            ):
                raise BundleError(_(
                    "Both the ZFS pools {self} and {other} on node {node} "
                    "try to use {devices} as the underlying storage device"
                ).format(
                    self=self.name,
                    other=item.name,
                    node=self.node.name,
                    devices=_(" and ").join(
                        set(item.devices_used).intersection(set(self.devices_used)),
                    ),
                ))

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if 'config' not in attributes.get('when_creating', {}):
            raise BundleError(_(
                "{item} on node {node}: required option 'config' missing"
            ).format(
                item=item_id,
                node=bundle.node.name,
            ))
        elif not isinstance(attributes['when_creating']['config'], list):
            raise BundleError(_(
                "{item} on node {node}: option 'config' must be a list"
            ).format(
                item=item_id,
                node=bundle.node.name,
            ))

        for config in attributes['when_creating']['config']:
            if config.get('type', None) not in {
                None,
                'mirror',
                'raidz',
                'raidz2',
                'raidz3',
                'cache',
                'log',
            }:
                raise BundleError(_(
                    "{item} on node {node} has invalid type '{type}', "
                    "must be one of (unset), 'mirror', 'raidz', 'raidz2', "
                    "'raidz3', 'cache', 'log'"
                ).format(
                    item=item_id,
                    node=bundle.node.name,
                    type=config['type'],
                ))

            if not config.get('devices', set()):
                raise BundleError(_(
                    "{item} on node {node} uses no devices!"
                ).format(
                    item=item_id,
                    node=bundle.node.name,
                ))

            if (
                config.get('type') == 'log' and
                len(config['devices']) not in (1, 2)
            ):
                raise BundleError(_(
                    "{item} on node {node} type 'log' must use exactly "
                    "one or two devices"
                ).format(
                    item=item_id,
                    node=bundle.node.name,
                ))



from base64 import b64encode

from bcrypt import hashpw as bcrypt_hashpw

# This is the current (2026) default cost factor in the underlying Rust
# library. (The Python wrapper doesn't export this symbol, so we have to
# copy the value here.)
_DEFAULT_BCRYPT_COST = 12

# bcrypt salts are 16 random bytes encoded by `b64encode_bcrypt()`.
_DEFAULT_BCRYPT_SALT = "oo2ahgheen9Tei0IeJohTO"


def b64encode_bcrypt(payload_bytes):
    payload_b64 = b64encode(payload_bytes).decode('ASCII')

    # This is the standard base64 alphabet:
    alphabet_standard = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'

    # Unfortunately, bcrypt uses a different one:
    # https://github.com/marshallpierce/rust-base64/blob/bc91a05e981345b7da3a5190cd3963eec6e98eb6/src/alphabet.rs#L187
    alphabet_bcrypt = './ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'

    # We must now translate from the first alphabet to the second one.
    t = str.maketrans(alphabet_standard, alphabet_bcrypt)
    payload_64_bcrypt = payload_b64.translate(t)

    # The padding must also be stripped off.
    return payload_64_bcrypt.rstrip('=')


def bcrypt(
    payload,
    encoding='UTF-8',
    cost=None,
    salt=None,
):
    """
    Returns a crypt line using the bcrypt algorithm (`2b`).

    These lines are useful for password hashes as used in UNIX or web
    servers.

    -   `encoding`: `payload` will be encoded using this encoding.
    -   `cost`: Use this cost factor, or `None` for bcrypt's default.
    -   `salt`: Must be a valid bcrypt salt, or `None` for bcrypt's default.
    """

    if cost is None:
        cost = _DEFAULT_BCRYPT_COST

    if salt is None:
        salt = _DEFAULT_BCRYPT_SALT

    # The bcrypt lib calls this "salt", but it's more than that, it also
    # includes the "2b" prefix and the cost factor.
    config = f'$2b${cost}${salt}'.encode('ASCII')

    return bcrypt_hashpw(payload.encode(encoding), config).decode(
        'ASCII'
    )



from os import makedirs
from os.path import join

from bundlewrap.utils.testing import make_repo, run


def test_not_a_repo_test(tmpdir):
    assert run("bw nodes", path=str(tmpdir))[2] == 1


def test_subdir_invocation(tmpdir):
    make_repo(tmpdir, nodes={"node1": {}})
    stdout, stderr, rcode = run("bw nodes", path=join(str(tmpdir), "bundles"))
    assert stdout == b"node1\n"
    assert stderr == b""
    assert rcode == 0


def test_toml_node_exists_in_new_repo(tmpdir):
    run("bw repo create", path=str(tmpdir))
    makedirs(join(tmpdir, "nodes"))
    with open(join(tmpdir, "nodes", "tomlnode.toml"), 'w') as f:
        f.write("")
    stdout, stderr, rcode = run("bw nodes", path=str(tmpdir))
    assert b"tomlnode" in stdout
    assert stderr == b""
    assert rcode == 0


def test_toml_group_exists_in_new_repo(tmpdir):
    run("bw repo create", path=str(tmpdir))
    makedirs(join(tmpdir, "groups"))
    with open(join(tmpdir, "groups", "tomlgroup.toml"), 'w') as f:
        f.write("")
    stdout, stderr, rcode = run("bw groups", path=str(tmpdir))
    assert b"tomlgroup" in stdout
    assert stderr == b""
    assert rcode == 0




from collections import defaultdict

from .deps import (
    prepare_dependencies,
    remove_item_dependents,
    remove_dep_from_items,
    resolve_selector,
    split_items_without_deps,
)
from .exceptions import NoSuchItem, MetadataUnavailable
from .utils.text import mark_for_translation as _
from .utils.ui import io


class BaseQueue:
    def __init__(self, node):
        self.items_with_deps = prepare_dependencies(node)
        self.items_without_deps = set()
        self._split()
        self.pending_items = set()

    def _split(self):
        self.items_with_deps, self.items_without_deps = \
            split_items_without_deps(self.all_items)

    @property
    def all_items(self):
        return self.items_with_deps | self.items_without_deps


class ItemQueue(BaseQueue):
    def __init__(self, node):
        super().__init__(node)

        # Optional sanity check.
        self.item_types_with_blockers = set()

    def item_failed(self, item):
        """
        Called when an item could not be fixed. Yields all items that
        have been skipped as a result by cascading.
        """
        for skipped_item in self.item_skipped(item):
            yield skipped_item

    def item_fixed(self, item):
        """
        Called when an item has successfully been fixed.
        """
        self.item_ok(item)
        self._fire_triggers_for_item(item)

    def item_ok(self, item):
        """
        Called when an item didn't need to be fixed.
        """
        self.pending_items.remove(item)
        # if an item is applied successfully, all dependencies on it can
        # be removed from the remaining items
        self.items_with_deps = remove_dep_from_items(
            self.items_with_deps,
            item,
        )
        self._split()

    def item_skipped(self, item):
        """
        Called when an item has been skipped. Yields all items that have
        been skipped as a result by cascading.
        """
        self.pending_items.remove(item)
        if item.cascade_skip:
            # if an item fails or is skipped, all items that depend on
            # it shall be removed from the queue
            self.items_with_deps, skipped_items = remove_item_dependents(
                self.items_with_deps,
                item,
            )
            for skipped_item in skipped_items:
                yield skipped_item
        else:
            self.items_with_deps = remove_dep_from_items(
                self.items_with_deps,
                item,
            )
        self._split()

    def items_without_deps_runnable(self):
        runnable_items = set()
        running_item_types = set([i.ITEM_TYPE_NAME for i in self.pending_items])

        for item in self.items_without_deps:
            add_this_item = True
            for item_blocked_for in item.block_concurrent(item.node.os, item.node.os_version):
                # Optional sanity check.
                #
                # Keep track of item types that have blockers. We can
                # later use this to do a sanity check: Was there a bug
                # and did we accidentally run blocked items after all?
                #
                # Note that this does NOT catch all cases that are
                # theoretically possible. It only catches things like
                # pkg_apt where only one item of that exact type can be
                # running.
                self.item_types_with_blockers.add(item.ITEM_TYPE_NAME)

                if item_blocked_for in running_item_types:
                    add_this_item = False
                    break

            if add_this_item:
                runnable_items.add(item)

        return runnable_items

    def pop(self):
        """
        Gets the next item available for processing and moves it into
        self.pending_items. Will raise KeyError if no item is
        available.
        """
        runnable_items = self.items_without_deps_runnable()

        if not runnable_items:
            raise MetadataUnavailable()

        item = runnable_items.pop()
        self.items_without_deps.remove(item)

        self.pending_items.add(item)

        # Optional sanity check.
        item_types_running = defaultdict(int)
        for i in self.pending_items:
            item_types_running[i.ITEM_TYPE_NAME] += 1
        for it in self.item_types_with_blockers:
            if item_types_running[it] > 1:
                raise Exception(f'BUG! More than one {it} running!')

        return item

    def _fire_triggers_for_item(self, item):
        for triggered_item_id in item.triggers:
            try:
                for triggered_item in resolve_selector(triggered_item_id, self.all_items):
                    triggered_item.has_been_triggered = True
            except (NoSuchItem, ValueError):
                io.debug(_(
                    "{item} tried to trigger {triggered_item}, "
                    "but it wasn't available. It must have been skipped previously."
                ).format(
                    item=item.id,
                    triggered_item=triggered_item_id,
                ))


class ItemTestQueue(BaseQueue):
    """
    A simpler variation of ItemQueue that is used by `bw test` to check
    for circular dependencies.
    """
    def pop(self):
        item = self.items_without_deps.pop()
        self.items_with_deps = remove_dep_from_items(self.items_with_deps, item)
        self._split()
        return item



from shlex import quote

from bundlewrap.items import Item


class ZFSDataset(Item):
    """
    Creates ZFS datasets and manages their options.
    """
    BUNDLE_ATTRIBUTE_NAME = "zfs_datasets"
    REJECT_UNKNOWN_ATTRIBUTES = False
    ITEM_TYPE_NAME = "zfs_dataset"

    def __repr__(self):
        return f"<ZFSDataset name:{self.name} {' '.join(f'{k}:{v}' for k,v in self.attributes.items())}>"

    def __create(self, path, options):
        option_list = []
        for option, value in sorted(options.items()):
            # We must exclude the 'mounted' property here because it's a
            # read-only "informational" property.
            if option != 'mounted' and value is not None:
                option_list.append("-o {}={}".format(quote(option), quote(value)))
        option_args = " ".join(option_list)

        self.run(
            "zfs create {} {}".format(
                option_args,
                quote(path),
            ),
            may_fail=True,
        )

        if options['mounted'] == 'no':
            self.__set_option(path, 'mounted', 'no')

    def __does_exist(self, path):
        status_result = self.run(
            "zfs list {}".format(quote(path)),
            may_fail=True,
        )
        return status_result.return_code == 0

    def __get_option(self, path, option):
        cmd = "zfs get -Hp -o value {} {}".format(quote(option), quote(path))
        # We always expect this to succeed since we don't call this function
        # if we have already established that the dataset does not exist.
        status_result = self.run(cmd)
        return status_result.stdout.decode('utf-8').strip()

    def __set_option(self, path, option, value):
        if option == 'mounted':
            # 'mounted' is a read-only property that can not be altered by
            # 'set'. We need to call 'zfs mount tank/foo'.
            self.run(
                "zfs {} {}".format(
                    "mount" if value == 'yes' else "unmount",
                    quote(path),
                ),
                may_fail=True,
            )
        else:
            self.run(
                "zfs set {}={} {}".format(
                    quote(option),
                    quote(value),
                    quote(path),
                ),
                may_fail=True,
            )

    @property
    def expected_state(self):
        state = {}
        for option, value in self.attributes.items():
            if option == 'mountpoint' and value is None:
                value = "none"
            if value is not None:
                state[option] = value

        state['mounted'] = 'no' if state.get('mountpoint') in (None, "none") else 'yes'
        return state

    def fix(self, status):
        if status.must_be_created:
            self.__create(self.name, status.expected_state)
        else:
            for option in status.keys_to_fix:
                self.__set_option(self.name, option, status.expected_state[option])

    def get_auto_attrs(self, items):
        pool = self.name.split("/")[0]
        pool_item = "zfs_pool:{}".format(pool)
        needs = set()

        for item in items:
            if item.ITEM_TYPE_NAME == "zfs_pool" and item.name == pool:
                # Add dependency to the pool this dataset resides on.
                needs.add(pool_item)
            elif (
                item.ITEM_TYPE_NAME == "zfs_dataset" and
                self.name != item.name
            ):
                # Find all other datasets that are parents of this
                # dataset.
                # XXX Could be optimized by finding the "largest"
                # parent only.
                if self.name.startswith(item.name + "/"):
                    needs.add(item.id)
                elif (
                    self.attributes.get('mountpoint')
                    and item.attributes.get('mountpoint')
                    and self.attributes['mountpoint'] != item.attributes['mountpoint']
                    and self.attributes['mountpoint'].startswith(item.attributes['mountpoint'])
                ):
                    needs.add(item.id)

        return {'needs': needs}

    @property
    def actual_state(self):
        if not self.__does_exist(self.name):
            return None

        state = {}
        for option in self.attributes:
            state[option] = self.__get_option(self.name, option)

        state['mounted'] = self.__get_option(self.name, 'mounted')
        return state



from os.path import join

from bundlewrap.utils.testing import make_repo, run


def test_satisfied(tmpdir):
    make_repo(
        tmpdir,
    )
    with open(join(str(tmpdir), "requirements.txt"), 'w') as f:
        f.write(
"""
bundlewrap
""")
    stdout, stderr, rcode = run("bw nodes", path=str(tmpdir))
    assert stderr == b""
    assert rcode == 0


def test_missing(tmpdir):
    make_repo(
        tmpdir,
    )
    with open(join(str(tmpdir), "requirements.txt"), 'w') as f:
        f.write(
"""
somepackagewhichisneverinstalled
""")
    stdout, stderr, rcode = run("bw nodes", path=str(tmpdir))
    assert stderr != b""
    assert rcode != 0


def test_wrong_version(tmpdir):
    make_repo(
        tmpdir,
    )
    with open(join(str(tmpdir), "requirements.txt"), 'w') as f:
        f.write(
"""
bundlewrap==1.0
""")
    stdout, stderr, rcode = run("bw nodes", path=str(tmpdir))
    assert stderr != b""
    assert rcode != 0


def test_marker(tmpdir):
    make_repo(
        tmpdir,
    )
    with open(join(str(tmpdir), "requirements.txt"), 'w') as f:
        f.write(
"""
bundlewrap==1.0 ; python_version<'1.0'
""")
    stdout, stderr, rcode = run("bw nodes", path=str(tmpdir))
    assert stderr == b""
    assert rcode == 0



from os.path import join
from sys import exit

from ..utils.text import mark_for_translation as _, red
from ..utils.ui import io


OPERATIONS = (
    'bytes',
    'decrypt',
    'encrypt',
    'human',
    'password',
)


def get_operation(args):
    opcount = 0
    selected_op = None
    for op in OPERATIONS:
        if args[op]:
            selected_op = op
            opcount += 1
    if opcount > 1:
        io.stdout(_("{x} More than one operation selected").format(x=red("!!!")))
        exit(1)
    elif opcount == 0:
        return 'password'
    return selected_op


def bw_pw(repo, args):
    if args['length'] < 1:
        io.stdout(_("{x} length must be > 1").format(x=red("!!!")))
        exit(1)

    op = get_operation(args)

    if op == 'bytes':
        io.stdout(repo.vault.random_bytes_as_base64_for(
            args['string'],
            key=args['key'] or 'generate',
            length=args['length'],
        ).value)

    elif op == 'decrypt':
        if args['file']:
            content = repo.vault.decrypt_file(
                args['string'],
                key=args['key'],
                binary=True,
            ).value
            with open(join(repo.data_dir, args['file']), 'wb') as f:
                f.write(content)
        else:
            try:
                key, cryptotext = args['string'].split("$", 1)
            except ValueError:
                cryptotext = args['string']
                key = args['key'] or 'encrypt'
            io.stdout(repo.vault.decrypt(
                cryptotext,
                key=key,
            ).value)

    elif op == 'encrypt':
        if args['file']:
            repo.vault.encrypt_file(
                args['string'],
                args['file'],
                key=args['key'] or 'encrypt',
            )
        else:
            io.stdout(repo.vault.encrypt(
                args['string'],
                key=args['key'] or 'encrypt',
            ))

    elif op == 'human':
        io.stdout(repo.vault.human_password_for(
            args['string'],
            key=args['key'] or 'generate',
        ).value)

    elif op == 'password':
        io.stdout(repo.vault.password_for(
            args['string'],
            key=args['key'] or 'generate',
            length=args['length'],
        ).value)



"""
Note that modules in this package have to use absolute imports because
Repository.item_classes loads them as files.
"""
from copy import copy
from datetime import datetime
from inspect import cleandoc
from os.path import join
from textwrap import TextWrapper

from bundlewrap.exceptions import (
    BundleError,
    FaultUnavailable,
    ItemDependencyError,
    ItemSkipped,
)
from bundlewrap.operations import run_local
from bundlewrap.utils import cached_property, Fault
from bundlewrap.utils.dicts import dict_to_text, diff_dict, hash_state_dict, validate_state_dict
from bundlewrap.utils.text import blue, bold, green, italic, red, wrap_question
from bundlewrap.utils.text import force_text, mark_for_translation as _
from bundlewrap.utils.ui import io

ALLOWED_ITEM_AUTO_ATTRIBUTES = {
    'after',
    'before',
    'needed_by',
    'needs',
}
BUILTIN_ITEM_ATTRIBUTES = {
    'after': set(),
    'before': set(),
    'cascade_skip': None,
    'comment': None,
    'needed_by': set(),
    'needs': set(),
    'preceded_by': set(),
    'precedes': set(),
    'error_on_missing_fault': False,
    'skip': False,
    'tags': set(),
    'triggered': False,
    'triggered_by': set(),
    'triggers': set(),
    'unless': "",
    'when_creating': {},
    'canned_actions_inherit_tags': True,
}

wrapper = TextWrapper(
    break_long_words=False,
    break_on_hyphens=False,
    expand_tabs=False,
    replace_whitespace=False,
)


def format_comment(comment):
    result = "\n\n"
    for line in wrapper.wrap(cleandoc(comment)):
        for inlineline in line.split("\n"):
            result += "{} {}\n".format(bold("#"), italic(inlineline))
    return result


def keys_to_fix(expected_state, actual_state):
    if expected_state is None:
        return set()
    if actual_state is None:
        return set(expected_state.keys())
    differing_keys = set()
    for key, value in expected_state.items():
        if value != actual_state[key]:
            differing_keys.add(key)
    return differing_keys


class ItemStatus:
    """
    Holds information on a particular Item such as whether it needs
    fixing and what's broken.
    """

    def __init__(self, expected_state, actual_state):
        self.expected_state = expected_state
        self.actual_state = actual_state
        self.keys_to_fix = set()
        self.must_be_deleted = (self.actual_state is not None and self.expected_state is None)
        self.must_be_created = (self.expected_state is not None and self.actual_state is None)
        if not self.must_be_deleted and not self.must_be_created:
            self.keys_to_fix = keys_to_fix(expected_state, actual_state)

    def __repr__(self):
        return "<ItemStatus correct:{}>".format(self.correct)

    @property
    def correct(self):
        return not self.must_be_deleted and not self.must_be_created and not bool(self.keys_to_fix)


def make_normalize(attribute_default):
    """
    This is to ensure you can pass filter() results and such in place of
    lists and have them converted to the proper type automatically.
    """
    if type(attribute_default) in (dict, list, set, tuple):
        def normalize(attribute_value):
            if attribute_value is None:
                return attribute_value
            else:
                return type(attribute_default)(attribute_value)

        return normalize
    else:
        return copy


class Item:
    """
    A single piece of configuration (e.g. a file, a package, a service).
    """
    BUNDLE_ATTRIBUTE_NAME = None
    ITEM_ATTRIBUTES = {}
    ITEM_TYPE_NAME = None
    REJECT_UNKNOWN_ATTRIBUTES = True
    REQUIRED_ATTRIBUTES = []
    SKIP_REASON_CMDLINE = 1
    SKIP_REASON_DEP_FAILED = 2
    SKIP_REASON_FAULT_UNAVAILABLE = 3
    SKIP_REASON_INTERACTIVE = 4
    SKIP_REASON_INTERACTIVE_ONLY = 5
    SKIP_REASON_NO_TRIGGER = 6
    SKIP_REASON_SOFTLOCK = 7
    SKIP_REASON_UNLESS = 8
    SKIP_REASON_DEP_SKIPPED = 9
    SKIP_REASON_ATTR = 10
    SKIP_REASON_DESC = {
        SKIP_REASON_CMDLINE: _("cmdline"),
        SKIP_REASON_DEP_FAILED: _("dependency failed"),
        SKIP_REASON_FAULT_UNAVAILABLE: _("Fault unavailable"),
        SKIP_REASON_INTERACTIVE: _("declined interactively"),
        SKIP_REASON_INTERACTIVE_ONLY: _("interactive only"),
        SKIP_REASON_NO_TRIGGER: _("not triggered"),
        SKIP_REASON_SOFTLOCK: _("soft locked"),
        SKIP_REASON_UNLESS: _("unless"),
        SKIP_REASON_DEP_SKIPPED: _("dependency skipped"),
        SKIP_REASON_ATTR: _("attribute"),
    }
    STATUS_OK = 1
    STATUS_FIXED = 2
    STATUS_FAILED = 3
    STATUS_SKIPPED = 4
    STATUS_ACTION_SUCCEEDED = 5
    WHEN_CREATING_ATTRIBUTES = {}

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        """
        Return a list of item types that cannot be applied in parallel
        with this item type.
        """
        return []

    def __init__(
        self,
        bundle,
        name,
        attributes,
        skip_validation=False,
        skip_name_validation=False,
    ):
        self.attributes = {}
        self.bundle = bundle
        self.has_been_triggered = False
        self.item_dir = join(bundle.bundle_dir, self.BUNDLE_ATTRIBUTE_NAME)
        self.item_data_dir = join(bundle.bundle_data_dir, self.BUNDLE_ATTRIBUTE_NAME)
        self.name = name
        self.node = bundle.node
        self.when_creating = {}
        self._command_results = []
        self._faults_missing_for_attributes = set()
        self._precedes_items = set()

        if not skip_validation:
            if not skip_name_validation:
                self._validate_name(bundle, name)
                self.validate_name(bundle, name)
            self._validate_attribute_names(bundle, self.id, attributes)
            self._validate_required_attributes(bundle, self.id, attributes)
            self.validate_attributes(bundle, self.id, attributes)

        try:
            attributes = self.patch_attributes(attributes)
        except FaultUnavailable:
            self._faults_missing_for_attributes.add(_("unknown"))

        for attribute_name, attribute_default in BUILTIN_ITEM_ATTRIBUTES.items():
            normalize = make_normalize(attribute_default)
            try:
                setattr(self, attribute_name, force_text(normalize(attributes.get(
                    attribute_name,
                    copy(attribute_default),
                ))))
            except FaultUnavailable:
                self._faults_missing_for_attributes.add(attribute_name)
                setattr(self, attribute_name, BUILTIN_ITEM_ATTRIBUTES[attribute_name])

        for attribute_name, attribute_default in self.ITEM_ATTRIBUTES.items():
            if attribute_name not in BUILTIN_ITEM_ATTRIBUTES:
                normalize = make_normalize(attribute_default)
                try:
                    self.attributes[attribute_name] = force_text(normalize(attributes.get(
                        attribute_name,
                        copy(attribute_default),
                    )))
                except FaultUnavailable:
                    self._faults_missing_for_attributes.add(attribute_name)

        for attribute_name, attribute_default in self.WHEN_CREATING_ATTRIBUTES.items():
            normalize = make_normalize(attribute_default)
            try:
                self.when_creating[attribute_name] = force_text(normalize(
                    attributes.get('when_creating', {}).get(
                        attribute_name,
                        copy(attribute_default),
                    )
                ))
            except FaultUnavailable:
                self._faults_missing_for_attributes.add('when_creating/' + attribute_name)

        if not self.REJECT_UNKNOWN_ATTRIBUTES:
            for key, value in attributes.items():
                if key in BUILTIN_ITEM_ATTRIBUTES:
                    continue
                if isinstance(value, Fault):
                    try:
                        value = value.value
                    except FaultUnavailable:
                        self._faults_missing_for_attributes.add(key)
                        continue
                self.attributes.setdefault(key, value)

        if self.cascade_skip is None:
            self.cascade_skip = not (self.skip or self.triggered or self.unless)

        if self.id in self.triggers:
            raise BundleError(_(
                "item {item} in bundle '{bundle}' can't trigger itself"
            ).format(
                bundle=self.bundle.name,
                item=self.id,
            ))

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __lt__(self, other):
        return self.id < other.id

    def __str__(self):
        return self.id

    def __repr__(self):
        return "<Item {}>".format(self.id)

    def _check_loopback_dependency(self):
        """
        Alerts the user if they have an item depend on itself.
        """
        if (
            self.id in self.after or
            self.id in self.before or
            self.id in self.needs or
            self.id in self.needed_by
        ):
            raise ItemDependencyError(_(
                "'{item}' in bundle '{bundle}' on node '{node}' cannot depend on itself"
            ).format(
                item=self.id,
                bundle=self.bundle.name,
                node=self.node.name,
            ))

    @cached_property
    def cached_expected_state(self):
        if self._faults_missing_for_attributes:
            self._raise_for_faults()

        expected_state = self.expected_state
        try:
            validate_state_dict(expected_state)
        except ValueError as e:
            raise ValueError(_(
                "{item} from bundle '{bundle}' returned invalid expected_state: {msg}"
            ).format(
                bundle=self.bundle.name,
                item=self.id,
                msg=repr(e),
            ))
        return expected_state

    @cached_property
    def cached_actual_state(self):
        status = self.actual_state
        try:
            validate_state_dict(status)
        except ValueError as e:
            raise ValueError(_(
                "{item} from bundle '{bundle}' returned invalid status: {msg}"
            ).format(
                bundle=self.bundle.name,
                item=self.id,
                msg=repr(e),
            ))
        return status

    @cached_property
    def cached_status(self):
        return self.get_status()

    @cached_property
    def cached_unless_result(self):
        """
        Returns True if 'unless' wants to skip this item.
        """
        if self.unless and (self.ITEM_TYPE_NAME == 'action' or not self.cached_status.correct):
            with io.job(_("{node}  {bundle}  {item}  running 'unless' ...").format(
                bundle=bold(self.bundle.name),
                item=self.id,
                node=bold(self.node.name),
            )):
                unless_result = self.node.run(self.unless, may_fail=True)
                return unless_result.return_code == 0
        else:
            return False

    def _triggers_preceding_items(self, interactive=False):
        """
        Preceding items will execute this to figure out if they're
        triggered.
        """
        if self.cached_unless_result:
            # 'unless' says we don't need to run
            return False
        if self.ITEM_TYPE_NAME == 'action':
            # so we have an action where 'unless' says it must be run
            # but the 'interactive' attribute might still override that
            if self.attributes['interactive'] and not interactive:
                return False
            else:
                return True
        return not self.cached_status.correct

    def _raise_for_faults(self):
        raise FaultUnavailable(_(
            "{item} on {node} is missing faults "
            "for these attributes: {attrs} "
            "(most of the time this means you're missing "
            "a required key in your .secrets.cfg)"
        ).format(
            attrs=", ".join(sorted(self._faults_missing_for_attributes)),
            item=self.id,
            node=self.node.name,
        ))

    def _skip_with_soft_locks(self, mine, others):
        """
        Returns True/False depending on whether the item should be
        skipped based on the given set of locks.
        """
        for lock in mine:
            if self.covered_by_autoskip_selector(lock['items']):
                io.debug(_("{item} on {node} whitelisted by lock {lock}").format(
                    item=self.id,
                    lock=lock['id'],
                    node=self.node.name,
                ))
                return False
        for lock in others:
            if self.covered_by_autoskip_selector(lock['items']):
                io.debug(_("{item} on {node} blacklisted by lock {lock}").format(
                    item=self.id,
                    lock=lock['id'],
                    node=self.node.name,
                ))
                return True
        return False

    def _test(self):
        with io.job(_("{node}  {bundle}  {item}").format(
            bundle=bold(self.bundle.name),
            item=self.id,
            node=bold(self.node.name),
        )):
            if self._faults_missing_for_attributes:
                self._raise_for_faults()
            return self.test()

    @classmethod
    def _validate_attribute_names(cls, bundle, item_id, attributes):
        if not isinstance(attributes, dict):
            raise BundleError(_(
                "invalid item '{item}' in bundle '{bundle}': not a dict"
            ).format(
                item=item_id,
                bundle=bundle.name,
            ))

        if cls.REJECT_UNKNOWN_ATTRIBUTES:
            invalid_attributes = set(attributes.keys()).difference(
                set(cls.ITEM_ATTRIBUTES.keys()).union(
                    set(BUILTIN_ITEM_ATTRIBUTES.keys())
                ),
            )
            if invalid_attributes:
                raise BundleError(_(
                    "invalid attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
                ).format(
                    item=item_id,
                    bundle=bundle.name,
                    attrs=", ".join(invalid_attributes),
                ))

            invalid_attributes = set(attributes.get('when_creating', {}).keys()).difference(
                set(cls.WHEN_CREATING_ATTRIBUTES.keys())
            )
            if invalid_attributes:
                raise BundleError(_(
                    "invalid when_creating attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
                ).format(
                    item=item_id,
                    bundle=bundle.name,
                    attrs=", ".join(invalid_attributes),
                ))

    @classmethod
    def _validate_name(cls, bundle, name):
        if not name:
            raise BundleError(_(
                "invalid name for {type} in bundle '{bundle}': must not be empty string"
            ).format(
                bundle=bundle.name,
                type=cls.ITEM_TYPE_NAME,
            ))
        if ":" in name:
            raise BundleError(_(
                "invalid name for {type} in bundle '{bundle}': {name} (must not contain colon)"
            ).format(
                bundle=bundle.name,
                name=name,
                type=cls.ITEM_TYPE_NAME,
            ))

    @classmethod
    def _validate_required_attributes(cls, bundle, item_id, attributes):
        missing = []
        for attrname in cls.REQUIRED_ATTRIBUTES:
            if attrname not in attributes:
                missing.append(attrname)
        if missing:
            raise BundleError(_(
                "{item} in bundle '{bundle}' missing required attribute(s): {attrs}"
            ).format(
                item=item_id,
                bundle=bundle.name,
                attrs=", ".join(missing),
            ))

    def apply(
        self,
        autoskip_selector=(),
        autoonly_selector=(),
        my_soft_locks=(),
        other_peoples_soft_locks=(),
        interactive=False,
        interactive_default=True,
        show_diff=True,
    ):
        self.node.repo.hooks.item_apply_start(
            repo=self.node.repo,
            node=self.node,
            item=self,
        )
        status_code = None
        status_before = None
        status_after = None
        details = None
        start_time = datetime.now()

        for item in self._precedes_items:
            try:
                if item._triggers_preceding_items(interactive=interactive):
                    io.debug(_(
                        "preceding item {item} on {node} has been triggered by {other_item}"
                    ).format(item=self.id, node=self.node.name, other_item=item.id))
                    self.has_been_triggered = True
                    break
                else:
                    io.debug(_(
                        "preceding item {item} on {node} has NOT been triggered by {other_item}"
                    ).format(item=self.id, node=self.node.name, other_item=item.id))
            except FaultUnavailable:
                io.debug(_(
                    "preceding item {item} on {node} has NOT been triggered by {other_item}, "
                    "because this other item is missing Faults"
                ).format(item=self.id, node=self.node.name, other_item=item.id))

        if self.skip:
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_ATTR

        elif self.triggered and not self.has_been_triggered:
            io.debug(_(
                "skipping {item} on {node} because it wasn't triggered"
            ).format(item=self.id, node=self.node.name))
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_NO_TRIGGER

        elif not self.covered_by_autoonly_selector(autoonly_selector):
            io.debug(_(
                "autoonly does not match {item} on {node}"
            ).format(item=self.id, node=self.node.name))
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_CMDLINE

        elif self.covered_by_autoskip_selector(autoskip_selector):
            io.debug(_(
                "autoskip matches {item} on {node}"
            ).format(item=self.id, node=self.node.name))
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_CMDLINE

        elif self._skip_with_soft_locks(my_soft_locks, other_peoples_soft_locks):
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_SOFTLOCK

        elif self._faults_missing_for_attributes:
            if self.error_on_missing_fault:
                self._raise_for_faults()
            else:
                io.debug(_(
                    "skipping {item} on {node} because it is missing faults "
                    "for these attributes: {attrs} "
                    "(most of the time this means you're missing "
                    "a required key in your .secrets.cfg)"
                ).format(
                    attrs=", ".join(sorted(self._faults_missing_for_attributes)),
                    item=self.id,
                    node=self.node.name,
                ))
                status_code = self.STATUS_SKIPPED
                details = self.SKIP_REASON_FAULT_UNAVAILABLE

        else:
            try:
                status_before = self.cached_status
            except FaultUnavailable:
                if self.error_on_missing_fault:
                    self._raise_for_faults()
                else:
                    io.debug(_(
                        "skipping {item} on {node} because it is missing Faults "
                        "(most of the time this means you're missing "
                        "a required key in your .secrets.cfg)"
                    ).format(
                        item=self.id,
                        node=self.node.name,
                    ))
                    status_code = self.STATUS_SKIPPED
                    details = self.SKIP_REASON_FAULT_UNAVAILABLE
            else:
                if self.cached_unless_result:
                    io.debug(_(
                        "'unless' for {item} on {node} succeeded, not fixing"
                    ).format(item=self.id, node=self.node.name))
                    status_code = self.STATUS_SKIPPED
                    details = self.SKIP_REASON_UNLESS
                elif status_before.correct:
                    status_code = self.STATUS_OK
                elif show_diff or interactive:
                    if status_before.must_be_created:
                        expected_state = copy(status_before.expected_state)
                        expected_state.update(self.when_creating)
                        details = self.display_on_create(expected_state)
                    elif status_before.must_be_deleted:
                        details = self.display_on_delete(copy(status_before.actual_state))
                    else:
                        details = self.display_on_fix(
                            copy(status_before.expected_state),
                            copy(status_before.actual_state),
                            copy(status_before.keys_to_fix),
                        )

        if status_code is None:  # item not skipped or OK
            if not interactive:
                with io.job(_("{node}  {bundle}  {item}").format(
                    bundle=bold(self.bundle.name),
                    item=self.id,
                    node=bold(self.node.name),
                )):
                    self.fix(status_before)
            else:
                if status_before.must_be_created:
                    question_text = dict_to_text(details, value_color=green)
                    prompt = _("Create {}?").format(bold(self.id))
                elif status_before.must_be_deleted:
                    question_text = dict_to_text(details, value_color=red)
                    prompt = _("Delete {}?").format(bold(self.id))
                else:
                    display_expected_state, display_actual_state, display_keys_to_fix = details
                    question_text = diff_dict(
                        display_actual_state,
                        display_expected_state,
                        skip_missing_in_target=True,
                    )
                    prompt = _("Fix {}?").format(bold(self.id))
                if self.comment:
                    question_text += format_comment(self.comment)
                question = wrap_question(
                    self.id,
                    question_text,
                    prompt,
                    prefix="{x} {node} ".format(
                        node=bold(self.node.name),
                        x=blue("?"),
                    ),
                )
                answer = io.ask(
                    question,
                    interactive_default,
                    epilogue="{x} {node}".format(
                        node=bold(self.node.name),
                        x=blue("?"),
                    ),
                )
                if answer:
                    with io.job(_("{node}  {bundle}  {item}").format(
                        bundle=bold(self.bundle.name),
                        item=self.id,
                        node=bold(self.node.name),
                    )):
                        self.fix(status_before)
                else:
                    status_code = self.STATUS_SKIPPED
                    details = self.SKIP_REASON_INTERACTIVE

        if status_code is None:  # item not skipped or OK
            status_after = self.get_status(cached=False)
            status_code = self.STATUS_FIXED if status_after.correct else self.STATUS_FAILED

        self.node.repo.hooks.item_apply_end(
            repo=self.node.repo,
            node=self.node,
            item=self,
            duration=datetime.now() - start_time,
            status_code=status_code,
            status_before=status_before,
            status_after=status_after,
        )
        return (
            status_code,
            details,
            status_before.must_be_created if status_before else None,
            status_before.must_be_deleted if status_before else None,
        )

    def run_local(self, command, **kwargs):
        result = run_local(command, **kwargs)
        self._command_results.append({
            'command': command,
            'result': result,
        })
        return result

    def run(self, command, **kwargs):
        result = self.node.run(command, **kwargs)
        self._command_results.append({
            'command': command,
            'result': result,
        })
        return result

    @property
    def expected_state(self):
        """
        Return a dict describing the expected state of this item on the node.
        Returning `None` instead means that the item should not exist.

        MAY be overridden by subclasses.
        """
        return self.attributes

    def covered_by_autoskip_selector(self, autoskip_selector):
        """
        True if this item should be skipped based on the given selector
        (e.g. ("tag:foo", "bundle:bar")).
        """
        components = [c.strip() for c in autoskip_selector]
        if (
            "*" in components or
            self.id in components or
            "bundle:{}".format(self.bundle.name) in components or
            "{}:".format(self.ITEM_TYPE_NAME) in components
        ):
            return True
        for tag in self.tags:
            if "tag:{}".format(tag) in components:
                return True
        return False

    def covered_by_autoonly_selector(self, autoonly_selector, check_deps=True):
        """
        True if this item should NOT be skipped based on the given selector
        (e.g. ("tag:foo", "bundle:bar")).
        """
        if not autoonly_selector:
            return True
        components = [c.strip() for c in autoonly_selector]
        if (
            self.id in components or
            "bundle:{}".format(self.bundle.name) in components or
            "{}:".format(self.ITEM_TYPE_NAME) in components
        ):
            return True
        for tag in self.tags:
            if "tag:{}".format(tag) in components:
                return True
        if check_deps:
            for depending_item in self._incoming_needs:
                if (
                    depending_item.id in components or
                    "bundle:{}".format(depending_item.bundle.name) in components or
                    "{}:".format(depending_item.ITEM_TYPE_NAME) in components
                ):
                    return True
                for tag in depending_item.tags:
                    if "tag:{}".format(tag) in components:
                        return True
        return False

    def fix(self, status):
        """
        This is supposed to actually implement stuff on the target node.

        MUST be overridden by subclasses.
        """
        raise NotImplementedError()

    def get_auto_attrs(self, items):
        """
        Return a dict with any number of attributes. The respective
        sets will be merged with the user-supplied values. For example:

            return {
                'needs': {
                    'file:/foo',
                },
            }

        Note that only attributes from ALLOWED_ITEM_AUTO_ATTRIBUTES are
        allowed.
        """
        return {}

    def get_canned_actions(self):
        """
        Return a dictionary of action definitions (mapping action names
        to dicts of action attributes, as in bundles).

        MAY be overridden by subclasses.
        """
        return {}

    def get_status(self, cached=True):
        """
        Returns an ItemStatus instance describing the current status of
        the item on the actual node.
        """
        with io.job(_("{node}  {bundle}  {item}").format(
            bundle=bold(self.bundle.name),
            item=self.id,
            node=bold(self.node.name),
        )):
            if not cached:
                del self._cache['cached_actual_state']
            return ItemStatus(self.cached_expected_state, self.cached_actual_state)

    def hash(self):
        return hash_state_dict(self.cached_expected_state)

    @property
    def id(self):
        if self.ITEM_TYPE_NAME == 'action' and ":" in self.name:
            # canned actions don't have an "action:" prefix
            return self.name
        return "{}:{}".format(self.ITEM_TYPE_NAME, self.name)

    def verify(
        self,
        autoskip_selector=(),
        autoonly_selector=(),
    ):
        if not self.covered_by_autoonly_selector(autoonly_selector, check_deps=False):
            io.debug(_(
                "autoonly does not match {item} on {node}"
            ).format(item=self.id, node=self.node.name))
            raise ItemSkipped

        if self.covered_by_autoskip_selector(autoskip_selector):
            io.debug(_(
                "autoskip matches {item} on {node}"
            ).format(item=self.id, node=self.node.name))
            raise ItemSkipped

        if self.cached_status.must_be_created:
            display = self.display_on_create(copy(self.cached_status.expected_state))
        elif self.cached_status.must_be_deleted:
            display = self.display_on_delete(copy(self.cached_status.actual_state))
        else:
            display = self.display_on_fix(
                copy(self.cached_status.expected_state),
                copy(self.cached_status.actual_state),
                copy(self.cached_status.keys_to_fix),
            )
        return self.cached_unless_result, self.cached_status, display

    def display_on_create(self, expected_state):
        """
        Given an expected_state dict as implemented above, modify it to
        better suit interactive presentation when an item is created. If
        there are any when_creating attributes, they will be added to
        the expected_state before it is passed to this method.

        MAY be overridden by subclasses.
        """
        return expected_state

    def display_on_fix(self, expected_state, actual_state, keys):
        """
        Given expected_state and actual_state as implemented above, modify them to
        better suit interactive presentation. The keys parameter is a
        set of keys whose values differ between expected_state and actual_state.

        MAY be overridden by subclasses.
        """
        return (expected_state, actual_state, keys)

    def display_on_delete(self, actual_state):
        """
        Given an actual_state dict as implemented above, modify it to
        better suit interactive presentation when an item is deleted.

        MAY be overridden by subclasses.
        """
        return actual_state

    def patch_attributes(self, attributes):
        """
        Allows an item to preprocess the attributes it is initialized
        with. Returns the modified attributes dictionary.

        MAY be overridden by subclasses.
        """
        return attributes

    def preview(self):
        """
        Can return a preview of this item as a Unicode string.
        BundleWrap will NOT add a trailing newline.

        MAY be overridden by subclasses.
        """
        raise NotImplementedError()

    @property
    def actual_state(self):
        """
        Return a dict describing the actual state of this item on the node.
        Returning `None` instead means that the item does not exist on the node.

        For the item to validate as correct, the values for all keys in
        self.expected_state have to match this actual_state.

        MUST be overridden by subclasses.
        """
        raise NotImplementedError()

    def test(self):
        """
        Used by `bw repo test`. Should do as much as possible to detect
        what would become a runtime error during a `bw apply`. Files
        will attempt to render their templates for example.

        SHOULD be overridden by subclasses
        """
        pass

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        """
        Raises BundleError if something is amiss with the user-specified
        attributes.

        SHOULD be overridden by subclasses.
        """
        pass

    @classmethod
    def validate_name(cls, bundle, name):
        """
        Raise BundleError if the given name is not valid (e.g. contains
        invalid characters for this kind of item.

        MAY be overridden by subclasses.
        """
        pass



import enum
from collections import defaultdict
from os import makedirs
from os.path import dirname, exists, join
from sys import exit

from ..deps import prepare_dependencies
from ..exceptions import FaultUnavailable
from ..items import BUILTIN_ITEM_ATTRIBUTES
from ..utils.cmdline import get_item, get_node
from ..utils.dicts import state_dict_to_json
from ..utils.table import ROW_SEPARATOR, render_table
from ..utils.text import bold, green, mark_for_translation as _, red, yellow
from ..utils.ui import io, page_lines


class ItemRepresentation(enum.Enum):
    PREVIEW = 'preview'
    ATTRS = 'attrs'
    ACTUAL_STATE = 'actual_state'
    EXPECTED_STATE = 'expected_state'
    REPR = 'repr'


def bw_items(repo, args):
    """
    Implementation for the `bw items` command
    """
    node = get_node(repo, args['node'])
    if args['file_preview_path']:  # -w / --write-file-previews
        render_file_previews(node, args['file_preview_path'], args)

    elif args['item']:  # [ITEM] specified
        item = get_item(node, args['item'])
        if args['preview']:  # -f / --preview
            representation = ItemRepresentation.PREVIEW
        elif args['show_attrs']:  # -a / --attrs
            representation = ItemRepresentation.ATTRS
        elif args['show_actual_state']:  # --state
            representation = ItemRepresentation.ACTUAL_STATE
        elif args['show_repr']:  # --repr
            representation = ItemRepresentation.REPR
        else:
            representation = ItemRepresentation.EXPECTED_STATE

        show_single_item(node, item, representation, args)

    else:
        list_all_items(node, args)


def render_file_previews(node, file_preview_path, args):
    """
    Implementation for `bw items --write-file-previews`: Writes all file-items into a local directory.
    """
    if args['item']:
        io.stderr(_("{x} use --preview to preview single files").format(x=red("!!!")))
        exit(1)

    if exists(file_preview_path):
        io.stderr(_(
            "not writing to existing path: {path}"
        ).format(path=file_preview_path))
        exit(1)

    for item in sorted(node.items):
        if not item.id.startswith("file:"):
            continue
        if item.attributes['content_type'] == 'any':
            io.stderr(_(
                "{x} skipped {filename} (content_type 'any')"
            ).format(x=yellow("»"), filename=bold(item.name)))
            continue
        if item.attributes['content_type'] == 'binary':
            io.stderr(_(
                "{x} skipped {filename} (content_type 'binary')"
            ).format(x=yellow("»"), filename=bold(item.name)))
            continue
        if item.attributes['content_type'] == 'download':
            io.stderr(_(
                "{x} skipped {filename} (content_type 'download')"
            ).format(x=yellow("»"), filename=bold(item.name)))
            continue
        if item.attributes['delete']:
            io.stderr(_(
                "{x} skipped {filename} ('delete' attribute set)"
            ).format(x=yellow("»"), filename=bold(item.name)))
            continue
        try:
            write_file_preview(item, file_preview_path)
        except FaultUnavailable:
            io.stderr(_(
                "{x} skipped {path} (Fault unavailable)"
            ).format(x=yellow("»"), path=bold(item.name)))
        else:
            io.stdout(_(
                "{x} wrote {path}"
            ).format(
                x=green("✓"),
                path=bold(join(
                    file_preview_path,
                    item.name.lstrip("/"),
                )),
            ))


def write_file_preview(file_item, base_path):
    """
    Writes the content of a single file item to the given path.
    """

    # this might raise an exception, try it before creating anything
    content = file_item.content

    # write content to named file
    file_path = join(base_path, file_item.name.lstrip("/"))
    dir_path = dirname(file_path)
    if not exists(dir_path):
        makedirs(dir_path)
    with open(file_path, 'wb') as f:
        f.write(content)


def show_single_item(node, item, representation, args):
    """
    Implementation for all variants of `bw items NODE ITEM`.
    """
    if representation == ItemRepresentation.PREVIEW:
        show_single_item_preview(node, item)
        return

    data = {}
    if representation == ItemRepresentation.ATTRS:
        prepare_dependencies(node)
        for attribute in BUILTIN_ITEM_ATTRIBUTES:
            if args['attr'] and attribute != args['attr']:
                continue

            value = getattr(item, attribute)
            data[attribute] = value

    elif item.ITEM_TYPE_NAME == "action":
        # actions don't have a state and thus no state-dict
        data = item.attributes

    elif representation == ItemRepresentation.ACTUAL_STATE:
        data = item.actual_state

    elif representation == ItemRepresentation.EXPECTED_STATE:
        data = item.expected_state

    elif representation == ItemRepresentation.REPR:
        data = [repr(item)]
        format_data(data, args, table_headers=[_('item')])
        return

    # print attributes key-value
    format_data(data, args, table_headers=[_('attribute'), _('value')])


def show_single_item_preview(node, item):
    """
    Implementation of `bw items NODE ITEM --preview`: Writes the content of single file to stdout.
    """
    try:
        io.stdout(
            item.preview(),
            append_newline=False,
        )
    except NotImplementedError:
        io.stderr(_(
            "{x} cannot preview {item} on {node} (doesn't support previews)"
        ).format(x=red("!!!"), item=item.id, node=node.name))
        exit(1)
    except ValueError:
        io.stderr(_(
            "{x} cannot preview {item} on {node} (not available for this item config)"
        ).format(x=red("!!!"), item=item.id, node=node.name))
        exit(1)
    except FaultUnavailable:
        io.stderr(_(
            "{x} cannot preview {item} on {node} (Fault unavailable)"
        ).format(x=red("!!!"), item=item.id, node=node.name))
        exit(1)


def list_all_items(node, args):
    """
    Implementation of `bw items NODE`: Lists all items on a node.
    """
    show_repr = args['show_repr']
    if args['blame']:
        # items per bundles
        data = defaultdict(list)
        table_headers = [_("bundle name"), _("items")]
        for item in sorted(node.items):
            data[item.bundle.name].append(repr(item) if show_repr else item.id)

    else:
        # items list
        data = []
        table_headers = [_('items')]
        for item in sorted(node.items):
            data.append(repr(item) if show_repr else item.id)

    format_data(data, args, table_headers=table_headers)


def format_data(data, args, table_headers=None):
    """
    Formats a list or a dict (with scalar or list values) according to the requested format
    """
    if args['format_json']:
        io.stdout(state_dict_to_json(data, pretty=True))

    else:
        format_data_table(data, table_headers)


def format_data_table(data, table_headers):
    """
    Formats a list or a dict (with scalar or list values) as bw-style table
    """
    table = [
        [
            bold(header)
            for header in table_headers
        ]
    ]

    if isinstance(data, (list, set)):
        if isinstance(data, set):
            data = list(sorted(data))

        table.append(ROW_SEPARATOR)
        for v in data:
            table.append([v])

    elif isinstance(data, dict):
        for k, v in sorted(data.items()):  # iterate items in dict
            table.append(ROW_SEPARATOR)
            table.append([str(k), v])

    page_lines(render_table(table))



from os.path import join

from bundlewrap.utils.testing import make_repo, run


def test_empty(tmpdir):
    make_repo(tmpdir)
    stdout, stderr, rcode = run("bw hash", path=str(tmpdir))
    assert stdout == b"44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a\n"
    assert stderr == b""
    assert rcode == 0


def test_nondeterministic(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content_type': 'mako',
                            'content': "<% import random %>${random.randint(1, 9999)}",
                        },
                    },
                },
            },
        },
    )

    hashes = set()

    for i in range(3):
        stdout, stderr, rcode = run("bw hash", path=str(tmpdir))
        assert rcode == 0
        hashes.add(stdout.strip())

    assert len(hashes) > 1


def test_deterministic(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content': "${node.name}",
                            'group': None,  # BSD has a different default and we don't want to
                                            # deal with that here
                        },
                    },
                },
            },
        },
    )

    hashes = set()

    for i in range(3):
        stdout, stderr, rcode = run("bw hash", path=str(tmpdir))
        assert rcode == 0
        hashes.add(stdout.strip())

    assert len(hashes) == 1
    assert hashes.pop() == b"d971ecd41722ad194e1d0ea3a5edacbcdf3b73d91c7c32302ce04c10b0705a7e"


def test_dict(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'actions': {
                        "test": {
                            'command': "true",
                        },
                    },
                    'files': {
                        "/test": {
                            'content': "yes please",
                            'group': None,  # BSD has a different default and we don't want to
                                            # deal with that here
                        },
                    },
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -d", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"e5ce2bede00e9edffa8f907f3bd6ddabc10e139ffa6c50719f46504f56bb4b9b  node1\n"

    stdout, stderr, rcode = run("bw hash -d node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"867841c209f2d269c5d395c5d2da55473e40c8e44a4c4fb4696e27c3f3e80e71  file:/test\n"

    stdout, stderr, rcode = run("bw hash -d node1 file:/test", path=str(tmpdir))
    assert rcode == 0
    assert stdout == (
        b"content_hash\t780f94dcf198f04a230c62fb135297eb3c37c9eb26c84d635ade85202789124c\n"
        b"mode\t0644\n"
        b"owner\troot\n"
        b"type\tfile\n"
    )


def test_metadata_empty(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {},
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -m node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a\n"


def test_metadata_fault(tmpdir):
    make_repo(tmpdir)
    with open(join(str(tmpdir), "nodes.py"), 'w') as f:
        f.write("""
nodes = {
    'node1': {
        'metadata': {'foo': vault.password_for("testing")},
    },
    'node2': {
        'metadata': {'foo': vault.password_for("testing").value},
    },
    'node3': {
        'metadata': {'foo': "faCTT76kagtDuZE5wnoiD1CxhGKmbgiX"},
    },
    'node4': {
        'metadata': {'foo': "something else entirely"},
    },
}
""")
    stdout1, stderr, rcode = run("bw hash -m node1", path=str(tmpdir))
    assert stdout1 == b"dc8aeb53a7175923f49f13a11b86f112634ffba5448f62f0fae6c01d12e67e6b\n"
    assert stderr == b""
    assert rcode == 0
    stdout2, stderr, rcode = run("bw hash -m node2", path=str(tmpdir))
    assert stdout2 == stdout1
    assert stderr == b""
    assert rcode == 0
    stdout3, stderr, rcode = run("bw hash -m node3", path=str(tmpdir))
    assert stdout3 == stdout1
    assert stderr == b""
    assert rcode == 0
    stdout4, stderr, rcode = run("bw hash -m node4", path=str(tmpdir))
    assert stdout4 != stdout1
    assert stderr == b""
    assert rcode == 0


def test_metadata_nested_sort(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {
                    'nested': {
                        'one': True,
                        'two': False,
                        'three': 3,
                        'four': "four",
                        'five': None,
                    },
                },
            },
            "node2": {
                'metadata': {
                    'nested': {
                        'five': None,
                        'four': "four",
                        'one': True,
                        'three': 3,
                        'two': False,
                    },
                },
            },
        },
    )

    stdout1, stderr, rcode = run("bw hash -m node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout1 == b"170200f5473a004c65fe1a87a12acb780533d571c990356e8aa02af8f690a08b\n"

    stdout2, stderr, rcode = run("bw hash -m node2", path=str(tmpdir))
    assert rcode == 0
    assert stdout1 == stdout2


def test_metadata_repo(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {
                    'foo': 47,
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -m", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"ff615a780a9181299ede59c07aed913fc4c7db0691fac7e2fbf54a7881f9051d\n"


def test_metadata_repo_dict(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {
                    'foo': 47,
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -md", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"node1\t51a11939dfc74a376361fd5dae8f40ed27243ee9d1f9133a7aa2e76a1e920964\n"


def test_groups_repo(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
    )

    stdout, stderr, rcode = run("bw hash -g", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"4fa6b006dce70008589915bdd008036e35a847131cdd2da98e6be585306e9e78\n"


def test_groups_repo_dict(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
    )

    stdout, stderr, rcode = run("bw hash -dg", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"group1\ngroup2\n"


def test_groups(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {'members': ["node1", "node2"]},
            "group2": {'members': ["node3"]},
        },
        nodes={
            "node1": {},
            "node2": {},
            "node3": {},
        },
    )

    stdout, stderr, rcode = run("bw hash -g group1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"763ba79953798760239be7150fd656e131df56b24ebbd4daec7b7c66ef27f993\n"


def test_groups_dict(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
        nodes={
            "node1": {'groups': {"group1"}},
            "node2": {'groups': {"group1"}},
            "node3": {'groups': {"group2"}},
        },
    )

    stdout, stderr, rcode = run("bw hash -dg group1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"node1\nnode2\n"


def test_groups_node(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
        nodes={
            "node1": {'groups': {"group1"}},
            "node2": {'groups': {"group1"}},
            "node3": {'groups': {"group2"}},
        },
    )

    stdout, stderr, rcode = run("bw hash -g node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"af67656b03c0e90bd3f900de3bbed794811185202bf9f996d78e3d2e550fd477\n"


def test_groups_node_dict(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
        nodes={
            "node1": {'groups': {"group1"}},
            "node2": {'groups': {"group1"}},
            "node3": {'groups': {"group2"}},
        },
    )

    stdout, stderr, rcode = run("bw hash -dg node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"group1\n"



import csv
from io import StringIO
from os import environ

from .text import ansi_clean

ROW_SEPARATOR = object()

if environ.get("BW_TABLE_STYLE") == 'ascii':
    FRAME_TOP_LEFT = "+-"
    FRAME_TOP_COLUMN_SEPARATOR = "-+-"
    FRAME_TOP_RIGHT = "-+"
    FRAME_BOTTOM_LEFT = "+-"
    FRAME_BOTTOM_COLUMN_SEPARATOR = "-+-"
    FRAME_BOTTOM_RIGHT = "-+"
    FRAME_CENTER_LEFT = "+-"
    FRAME_CENTER_COLUMN_SEPARATOR = "-+-"
    FRAME_CENTER_RIGHT = "-+"
    FRAME_COLUMN_FILLER = "-"
    FRAME_COLUMN_WHITESPACE = " "
    FRAME_ROW_COLUMN_SEPARATOR_LEFT = "-| "
    FRAME_ROW_COLUMN_SEPARATOR_NONE = " | "
    FRAME_ROW_COLUMN_SEPARATOR_BOTH = "-+-"
    FRAME_ROW_COLUMN_SEPARATOR_RIGHT = " |-"
elif environ.get("BW_TABLE_STYLE") == 'grep':
    FRAME_TOP_LEFT = ""
    FRAME_TOP_COLUMN_SEPARATOR = ""
    FRAME_TOP_RIGHT = ""
    FRAME_BOTTOM_LEFT = ""
    FRAME_BOTTOM_COLUMN_SEPARATOR = ""
    FRAME_BOTTOM_RIGHT = ""
    FRAME_CENTER_LEFT = ""
    FRAME_CENTER_COLUMN_SEPARATOR = ""
    FRAME_CENTER_RIGHT = ""
    FRAME_COLUMN_FILLER = ""
    FRAME_COLUMN_WHITESPACE = ""
    FRAME_ROW_COLUMN_SEPARATOR_LEFT = "\t"
    FRAME_ROW_COLUMN_SEPARATOR_NONE = "\t"
    FRAME_ROW_COLUMN_SEPARATOR_BOTH = "\t"
    FRAME_ROW_COLUMN_SEPARATOR_RIGHT = "\t"
else:
    FRAME_TOP_LEFT = "╭─"
    FRAME_TOP_COLUMN_SEPARATOR = "─┬─"
    FRAME_TOP_RIGHT = "─╮"
    FRAME_BOTTOM_LEFT = "╰─"
    FRAME_BOTTOM_COLUMN_SEPARATOR = "─┴─"
    FRAME_BOTTOM_RIGHT = "─╯"
    FRAME_CENTER_LEFT = "├─"
    FRAME_CENTER_COLUMN_SEPARATOR = "─┼─"
    FRAME_CENTER_RIGHT = "─┤"
    FRAME_COLUMN_FILLER = "─"
    FRAME_COLUMN_WHITESPACE = " "
    FRAME_ROW_COLUMN_SEPARATOR_LEFT = "─┤ "
    FRAME_ROW_COLUMN_SEPARATOR_NONE = " │ "
    FRAME_ROW_COLUMN_SEPARATOR_BOTH = "─┼─"
    FRAME_ROW_COLUMN_SEPARATOR_RIGHT = " ├─"


def _column_widths_for_rows(rows):
    column_widths = [0 for column in rows[0]]
    for row in rows:
        if not isinstance(row, list) and not isinstance(row, tuple):
            continue
        for i, column in enumerate(row):
            if column == ROW_SEPARATOR:
                continue
            if isinstance(column, (list, tuple, set)):
                for value in column:
                    column_widths[i] = max(column_widths[i], len(ansi_clean(value)))
            else:
                column_widths[i] = max(column_widths[i], len(ansi_clean(column)))
    return column_widths


def _border_top(column_widths):
    result = FRAME_TOP_LEFT
    result += FRAME_TOP_COLUMN_SEPARATOR.join(
        [FRAME_COLUMN_FILLER * width for width in column_widths]
    )
    result += FRAME_TOP_RIGHT
    return result


def _border_center(column_widths):  # FIXME unused?
    result = FRAME_CENTER_LEFT
    result += FRAME_CENTER_COLUMN_SEPARATOR.join(
        [FRAME_COLUMN_FILLER * width for width in column_widths]
    )
    result += FRAME_CENTER_RIGHT
    return result


def _border_bottom(column_widths):
    result = FRAME_BOTTOM_LEFT
    result += FRAME_BOTTOM_COLUMN_SEPARATOR.join(
        [FRAME_COLUMN_FILLER * width for width in column_widths]
    )
    result += FRAME_BOTTOM_RIGHT
    return result


def _empty_row(row):
    for column_value in row:
        if column_value != ROW_SEPARATOR and column_value.strip():
            return False
    return True


def _nested_row(row, column_widths, alignments):
    repeat_literal_columns = environ.get("BW_TABLE_STYLE") == 'grep'
    for flat_row in _flatten_rows_with_lists(row, repeat_literal_columns):
        yield _row(flat_row, column_widths, alignments)


def _row(row, column_widths, alignments):
    result = ""
    columns = []
    for i, column_value in enumerate(row):
        alignment = alignments.get(i, 'left')

        if column_value == ROW_SEPARATOR:
            columns.append(ROW_SEPARATOR)
        elif alignment == 'right':
            columns.append(
                FRAME_COLUMN_WHITESPACE * (column_widths[i] - len(ansi_clean(column_value))) +
                str(column_value)
            )
        elif alignment == 'left':
            columns.append(
                str(column_value) +
                FRAME_COLUMN_WHITESPACE * (column_widths[i] - len(ansi_clean(column_value)))
            )
        elif alignment == 'center':
            prefix = int((column_widths[i] - len(ansi_clean(column_value))) / 2)
            suffix = (column_widths[i] - len(ansi_clean(column_value)) - prefix)
            columns.append(
                FRAME_COLUMN_WHITESPACE * prefix +
                str(column_value) +
                FRAME_COLUMN_WHITESPACE * suffix
            )
        else:
            raise NotImplementedError("no such alignment: {}".format(alignment))

    for i, column_value in enumerate(columns):
        if i == 0:
            fill_previous_column = False
        else:
            fill_previous_column = columns[i - 1] == ROW_SEPARATOR
        fill_this_column = column_value == ROW_SEPARATOR

        if fill_previous_column and fill_this_column:
            result += FRAME_ROW_COLUMN_SEPARATOR_BOTH
        elif fill_previous_column and not fill_this_column:
            result += FRAME_ROW_COLUMN_SEPARATOR_LEFT
        elif not fill_previous_column and fill_this_column:
            result += FRAME_ROW_COLUMN_SEPARATOR_RIGHT
        else:
            result += FRAME_ROW_COLUMN_SEPARATOR_NONE

        if fill_this_column:
            result += FRAME_COLUMN_FILLER * column_widths[i]
        else:
            result += column_value

    if fill_this_column:
        result += FRAME_ROW_COLUMN_SEPARATOR_LEFT
    else:
        result += FRAME_ROW_COLUMN_SEPARATOR_NONE

    return result[1:-1]  # strip exactly one whitespace character at each end


def _flatten_rows_with_lists(row, repeat_literal_columns):
    has_list_column = any(isinstance(cell, (list, tuple, set)) for cell in row)
    if not has_list_column:
        yield row
        return

    longest_list_column = max(len(cell) for cell in row if isinstance(cell, (list, tuple, set)))
    longest_list_column = max(longest_list_column, 1)
    is_first_flat_row = True

    for i in range(longest_list_column):
        flat_row = []
        for cell in row:
            if isinstance(cell, set):
                cell = list(sorted(cell))

            if isinstance(cell, (list, tuple)):
                if len(cell) == 0:
                    flat_row.append('[]')
                elif i < len(cell):
                    flat_row.append(str(cell[i]))
                else:
                    flat_row.append('')
            else:
                if is_first_flat_row or repeat_literal_columns:
                    flat_row.append(str(cell))
                else:
                    flat_row.append('')

        is_first_flat_row = False
        yield flat_row


def inline_list(cell):
    if isinstance(cell, set):
        return ",".join(sorted(set))
    elif isinstance(cell, (list, tuple)):
        return ",".join(cell)
    else:
        return cell


def render_table(rows, alignments=None):
    """
    Yields lines for a table.

    rows must be a list of lists of values, with the first row being
    considered the heading row. Alternatively, an entire row or
    individual cells can be set to ROW_SEPARATOR to turn it into a
    separator:

    rows = [
        ["heading1", "heading2"],
        ROW_SEPARATOR,
        ["value1", "value2"],
        ["value3", ROW_SEPARATOR],
    ]

    alignments is a dict mapping column indexes to 'left' or 'right'.
    """
    if not rows:
        return

    num_columns = len(rows[0])

    if environ.get("BW_TABLE_STYLE") == 'csv':
        yield from render_table_csv(rows)
    elif num_columns == 1:
        yield from render_table_single_column(rows)
    else:
        yield from render_table_beauty(alignments, rows)


def render_table_csv(rows):
    output = StringIO()
    writer = csv.writer(output)
    for row in rows:
        if row == ROW_SEPARATOR:
            continue
        writer.writerow([ansi_clean(inline_list(cell)) for cell in row])
    for line in output.getvalue().splitlines():
        yield line


def render_table_single_column(rows):
    for row in rows:
        if not row:
            yield ""
        elif row == ROW_SEPARATOR:
            continue
        else:
            yield str(row[0])


def render_table_beauty(alignments, rows):
    if alignments is None:
        alignments = {}
    column_widths = _column_widths_for_rows(rows)

    if environ.get("BW_TABLE_STYLE") != 'grep':
        yield _border_top(column_widths)

    for row_index, row in enumerate(rows):
        if row == ROW_SEPARATOR:
            if environ.get("BW_TABLE_STYLE") != 'grep':
                yield from _nested_row([ROW_SEPARATOR] * len(column_widths), column_widths, {})
        elif row_index == 0:
            # heading row ignores alignments
            yield from _nested_row(row, column_widths, {})
        elif environ.get("BW_TABLE_STYLE") != 'grep' or not _empty_row(row):
            yield from _nested_row(row, column_widths, alignments)

    if environ.get("BW_TABLE_STYLE") != 'grep':
        yield _border_bottom(column_widths)



# PYTHON_ARGCOMPLETE_OK

from cProfile import Profile
from os import environ
from os.path import abspath
from shlex import quote
from sys import argv, exit
from traceback import format_exc


from ..exceptions import NoSuchRepository, MissingRepoDependency
from ..repo import Repository
from ..utils.cmdline import suppress_broken_pipe_msg
from ..utils.text import force_text, mark_for_translation as _, red
from ..utils.ui import io
from .parser import build_parser_bw


@suppress_broken_pipe_msg
def main(*args, **kwargs):
    """
    Entry point for the 'bw' command line utility.

    The args and path parameters are used for integration tests.
    """
    if not args:
        args = argv[1:]

    text_args = [force_text(arg) for arg in args]

    parser_bw = build_parser_bw()
    pargs = parser_bw.parse_args(args)
    if not hasattr(pargs, 'func'):
        parser_bw.print_help()
        exit(2)
    if pargs.profile:
        profile = Profile()
        profile.enable()

    path = abspath(pargs.repo_path)
    io.debug_mode = pargs.debug
    io.activate()
    io.debug(_("invocation: {}").format(" ".join([force_text(arg) for arg in argv])))

    environ.setdefault('BW_ADD_HOST_KEYS', "1" if pargs.add_ssh_host_keys else "0")

    if len(text_args) >= 1 and (
        text_args[0] == "--version" or
        (len(text_args) >= 2 and text_args[0] == "repo" and text_args[1] == "create") or
        text_args[0] == "zen" or
        "-h" in text_args or
        "--help" in text_args
    ):
        # 'bw repo create' is a special case that only takes a path
        repo = path
    else:
        try:
            repo = Repository(path)
        except NoSuchRepository:
            io.stderr(_(
                "{x} {path} "
                "is not a BundleWrap repository."
            ).format(path=quote(abspath(pargs.repo_path)), x=red("!!!")))
            io.deactivate()
            exit(1)
        except MissingRepoDependency as exc:
            io.stderr(str(exc))
            io.deactivate()
            exit(1)
        except Exception:
            io.stderr(format_exc())
            io.deactivate()
            exit(1)

    # convert all string args into text
    ## hint from the future: not all args are indeed text and this code silently assumes, that force_text will keep them...
    text_pargs = {key: force_text(value) for key, value in vars(pargs).items()}

    try:
        pargs.func(repo, text_pargs)
    finally:
        io.deactivate()
        if pargs.profile:
            profile.disable()
            profile.dump_stats(pargs.profile)
